# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABCMeta
from datetime import datetime
from functools import cache
from types import MappingProxyType  # A read-only dict
from typing import Any, Iterable, Optional, Type

from sqlalchemy import JSON, inspect
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import (
    Mapped,
    MappedColumn,
    RelationshipDirection,
    declarative_base,
    declared_attr,
    mapped_column
)

from .exception import BadColumnError
from .relationship import InstanceRelationDict
from ..core import DataSourceError


class Model:
    """
    A model that can be converted to DataObject instances.

    The properties can be implemented in many different ways, but
    are necessary on every child class that is exposed to
    SqlDataSource.

    Relationships that start with an underscore (e.g. `User._tokens`)
    will be ignored by these methods.
    """

    @classmethod
    def get_table_name(cls) -> str:  # noqa
        """The name of the Model"""

    @classmethod
    def get_id_column_name(cls) -> str:  # noqa
        """
        The name of the column that serves as the "id".
        Override this classmethod to change.
        """

    @classmethod
    def get_column(cls, name: str) -> MappedColumn:
        """The (attribute) column for the given name."""

    @classmethod
    def get_to_one_relationship_config(cls) -> dict[str, str]:
        """
        The mapping of relationship names to tablenames for to-one relationships
        """

    @classmethod
    def get_to_many_relationship_config(cls) -> dict[str, str]:
        """
        The mapping of relationship names to tablenames for to-many relationships
        """

    @classmethod
    def get_attribute_types(cls) -> dict[str, type]:
        """
        The mapping of attribute names to their datatype in python
        """

    @classmethod
    def get_foreign_key_name(cls, relationship_name: str) -> str:
        """
        The name of the foreign key column for the given relationship name
        """

    @classmethod
    def get_all_foreign_key_names(cls) -> set[str]:
        """
        Returns only the names of columns which are foreign keys used in
        relationsips.
        """

    def before_commit(self, user_id: Optional[str] = None) -> None:
        """
        The method that is called by a `Database` instance before a commit
        """

    @classmethod
    def get_excluded_column_names(cls) -> list[str]:
        """
        The list of column names that both exist on the model, but
        shouldn't be automatically added to `Model().instance_attributes`.
        """

        return [
            cls.get_id_column_name()
        ]

    @property
    def instance_to_one_relations(self) -> dict[str, Optional[Model]]:
        """
        The mapping of relationship names to to-one relation rows
        """

    @property
    def instance_to_many_relations(self) -> dict[str, Iterable[Model]]:
        """
        The mapping of relationship names to to-many relation rows
        """

    @property
    def instance_id(self) -> Optional[str]:
        """The (potentially None) id of this model instance"""

    @property
    def instance_attributes(self) -> dict[str, Any]:
        """The Dict of attribute key to values"""


class InstanceToOneDict(
    InstanceRelationDict[Optional[Model]]
):
    """
    A useful concretion for the to-one instance relationship dict
    """

    @property
    def config(self) -> dict[str, str]:
        return self.source.get_to_one_relationship_config()


class InstanceToManyDict(
    InstanceRelationDict[Iterable[Model]]
):
    """
    A useful concretion for the to-many instance relationship dict
    """

    @property
    def config(self) -> dict[str, str]:
        return self.source.get_to_many_relationship_config()


class DefaultModel(Model):
    Log: Model
    """
    An inherited class that logs changes to its fields:
    - when
    - by whom
    """


def model_base() -> Type[DefaultModel]:
    """
    Creates a new base for Model classes that implement the Model ABC.
    """

    class ModelMeta(DeclarativeMeta, ABCMeta):
        pass

    DeclarativeBase = declarative_base(  # noqa N806
        metaclass=ModelMeta,
        type_annotation_map={
            dict: JSON,
            dict[str, Any]: JSON
        }
    )

    class ModelBase(DeclarativeBase, DefaultModel):
        """
        An ABC that implements the Model ABC, using reasonable defaults.

        Either:
        - use this class for reasonable default behaviour
        - inherit from both DeclarativeBase and Model, implementing the
        required abstract members, if custom behaviour is required.
        """

        __abstract__ = True

        __tablename__: str
        """The name of this table in the actual DB"""

        @classmethod
        def get_table_name(cls) -> str:
            return cls.__tablename__

        @classmethod
        def get_id_column_name(cls) -> str:
            return 'id'

        @classmethod
        def get_column(cls, name: str) -> MappedColumn:
            if name not in inspect(cls).mapper.attrs:
                raise BadColumnError(cls, name)
            return getattr(cls, name)

        @classmethod
        @cache
        def get_to_many_relationship_config(cls) -> MappingProxyType[str, str]:
            relationships = inspect(cls).relationships
            all_ = {
                r.key: cls.__get_relationship_target(r)
                for r in relationships
                if cls.__is_to_many_relationship(r)
            }
            return MappingProxyType({
                k: v for k, v in all_.items()
                if not k.startswith('_')
            })

        @classmethod
        @cache
        def get_to_one_relationship_config(cls) -> MappingProxyType[str, str]:
            relationships = inspect(cls).relationships
            all_ = {
                r.key: cls.__get_relationship_target(r)
                for r in relationships
                if cls.__is_to_one_relationship(r)
            }
            return MappingProxyType({
                k: v for k, v in all_.items()
                if not k.startswith('_')
            })

        @classmethod
        def get_foreign_key_name(cls, relationship_name: str) -> str:
            # TODO refactor `Database().get_to_one_relation()` to use this
            foreign_key = cls.__get_foreign_key(relationship_name)
            return foreign_key.name

        @classmethod
        @cache
        def get_attribute_types(cls) -> MappingProxyType[str, type]:
            names = cls.__get_attribute_names()
            columns = inspect(cls).columns
            return MappingProxyType({
                k: columns[k].type.python_type for k in names
            })

        @classmethod
        def get_id_attribute_type(cls) -> dict[str, type]:
            columns = inspect(cls).columns
            return columns[cls.get_id_column_name()].type.python_type

        @property
        def instance_to_one_relations(self) -> dict[str, Optional[Model]]:
            config = self.get_to_one_relationship_config()
            return self.__get_attributes_map(config)

        @property
        def instance_to_many_relations(self) -> dict[str, Iterable[Model]]:
            config = self.get_to_many_relationship_config()
            return self.__get_attributes_map(config)

        @property
        def instance_id(self) -> Optional[str]:
            id_key = self.get_id_column_name()
            id_val = getattr(self, id_key)
            return None if id_val is None else str(id_val)

        @property
        def instance_attributes(self) -> dict[str, Any]:
            names = self.__get_attribute_names()
            return self.__get_attributes_map(names)

        @classmethod
        def __get_foreign_key(
            cls,
            relationship_name: str
        ) -> MappedColumn:
            """
            Returns the local foreign key column given a relationship name
            """

            rel = inspect(cls).relationships[relationship_name]
            all_cols = tuple(rel.local_columns)
            if len(all_cols) != 1:
                raise NotImplementedError('Composite keys are not supported.')
            return all_cols[0]

        def __get_attributes_map(
            self,
            names: Iterable[str]
        ) -> dict[str, Any]:

            return {
                name: getattr(self, name) for name in names
            }

        @classmethod
        def __get_relationship_target(cls, relationship) -> str:
            return relationship.entity.tables[0].name

        @classmethod
        def __get_all_relationship_names(cls) -> list[str]:
            mapper = inspect(cls)
            return list(mapper.relationships.keys())

        @classmethod
        def __is_to_one_relationship(cls, relationship) -> bool:
            return relationship.direction == RelationshipDirection.MANYTOONE

        @classmethod
        def __is_to_many_relationship(cls, relationship) -> bool:
            return relationship.direction in (
                RelationshipDirection.ONETOMANY,
                RelationshipDirection.MANYTOMANY
            )

        @classmethod
        @cache
        def get_all_foreign_key_names(cls) -> frozenset[str]:
            """
            Returns only the names of columns which are foreign keys used in
            relationsips. i.e. To-one relationships.
            """

            col_to_attr = cls.__get_column_name_to_attribute_map()

            foreign_keys = set()
            for rel in inspect(cls).relationships:
                for col in rel.local_columns:
                    # If it is really a foreign key, get the name of the
                    # attribute in the model
                    if len(col.foreign_keys) > 0 and (attr_name := col_to_attr.get(col.name)):
                        foreign_keys.add(attr_name)
            return frozenset(foreign_keys)

        @classmethod
        def __get_column_name_to_attribute_map(cls) -> dict[str, str]:
            """
            Returns a dict mapping database column names to this model's
            attribute names.
            """

            return {
                col_prop.columns[0].name: col_prop.key
                for col_prop in inspect(cls).column_attrs
            }

        @classmethod
        @cache
        def __get_attribute_names(cls) -> tuple[str]:
            excluded = cls.get_excluded_column_names()
            mapper = inspect(cls)
            relationships = cls.__get_all_relationship_names()
            foreign_keys = cls.get_all_foreign_key_names()
            return tuple(
                k for k in mapper.attrs.keys()  # noqa: SIM118
                if k not in excluded
                and k not in relationships
                and k not in foreign_keys
            )

    class LogBase(ModelBase):
        """
        Logs changes to its columns. (by whom/when)
        """

        __abstract__ = True

        @declared_attr
        def modified_at(self) -> Mapped[datetime]:
            return mapped_column(nullable=True)

        @declared_attr
        def modified_by(self) -> Mapped[str]:
            return mapped_column(nullable=True)

        def before_commit(
            self,
            *,
            user_id: Optional[str]
        ) -> None:

            if user_id is None:
                self.__raise_unknown_user()
            self.__update_log_fields(user_id)

        def __update_log_fields(
            self,
            user_id: str
        ) -> None:

            self.modified_by = user_id
            self.modified_at = datetime.now()

        def __raise_unknown_user(self) -> None:
            raise DataSourceError(
                title='Unauthenticated',
                detail='No user identified for this change.',
                status_code=401
            )

    ModelBase.Log = LogBase

    return ModelBase



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class BioscanQcSpecimenToElasticSampleUpdateConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObjectUpdate]:
        if data_object is not None and data_object.id is not None:

            yield (None, {
                'sts_specimen.id': data_object.id,
                'sanger_qc_result': data_object.sanger_qc_result,
                'sanger_qc_description': data_object.sanger_qc_description})



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .elastic_datasource import *  # noqa
from .factory import create_elastic_datasource  # noqa
from .runtime_fields import *  # noqa



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re
from typing import Iterable

from dateutil.parser import parse as dateutil_parse

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class GritIssueToElasticIssueConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        status_changes = {
            self.__sanitise_attribute_name(sc['next_status']) + '_date': sc['end_date']
            for sc in data_object.status_changes
            if sc.get('next_status') is not None
        }
        sprints = self.__parse_sprint(data_object.sprint)

        sprint_attributes = {
            'number_of_sprints': len(sprints),
            'first_sprint_start_date': sprints[0]['startDate'] if sprints else None,
            'first_sprint_end_date': sprints[0]['endDate'] if sprints else None,
            'first_sprint_name': sprints[0]['name'] if sprints else None,
            'last_sprint_start_date': sprints[-1]['startDate'] if sprints else None,
            'last_sprint_end_date': sprints[-1]['endDate'] if sprints else None,
            'last_sprint_name': sprints[-1]['name'] if sprints else None,
        }

        yield self._data_object_factory(
            'issue',
            data_object.id,
            attributes={
                'issue_type': data_object.issue_type,
                'jira_project': data_object.project,
                'created': data_object.created,
                'status': data_object.status,
                'assignee_name': data_object.assignee.name if data_object.assignee else None,
                'original_story_points': data_object.original_story_points,
                'story_points': data_object.story_points,
                'original_estimate': data_object.original_estimate,
                'remaining_estimate': data_object.remaining_estimate
            } | status_changes | sprint_attributes
        )

    def __sanitise_attribute_name(self, name: str) -> str:
        return re.sub(r'\s+', '_', name.lower())

    def __parse_sprint(self, sprint: list | None) -> str:
        sprints = []
        if sprint is None:
            return sprints
        for s in sprint:
            match_end = re.search(r'endDate=([^,\]]+)', s)
            match_name = re.search(r'name=([^,\]]+)', s)
            match_start = re.search(r'startDate=([^,\]]+)', s)

            if match_end and match_name and match_start and match_start.group(1) != '<null>' \
                    and match_end.group(1) != '<null>':
                start_date = dateutil_parse(match_start.group(1)) if match_start.group(1) else None
                end_date = dateutil_parse(match_end.group(1)) if match_end.group(1) else None
                sprints.append({
                    'name': match_name.group(1),
                    'startDate': start_date,
                    'endDate': end_date
                })
        return sprints



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re


def sanitise_value(value, default=''):
    if value is None:
        value = ''
    if default != '' and value == '':
        value = default
    if isinstance(value, str):
        value = re.sub(r'\\n', '\n', value)
        value = re.sub(r'\\t', '\t', value)
        value = value.strip()
    return value



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any

from .action import Action
from ..core import DataSource


class UpsertAction(Action):
    """
    The central class for running local actions.
    """

    def __init__(self):
        super().__init__()

    def run(
        self,
        datasource: DataSource,
        ids: list[str],
        object_type: str,
        params: dict[str, Any] | None = None
    ) -> tuple[dict[str, bool], int]:

        data_objects = self.__convert_to_data_objects(
            datasource=datasource,
            ids=ids,
            object_type=object_type,
            params=params
        )

        try:
            datasource.upsert_batch(object_type=object_type, objects=data_objects)
            return {'success': True}, 200
        except Exception as e:
            return {'error': str(e)}, 500

    def __convert_to_data_objects(
        self,
        datasource: DataSource,
        ids: list[str],
        object_type: str,
        params: dict[str, Any] | None = None
    ) -> Any:
        CoreDataObject = datasource.data_object_factory # noqa N806

        for id_ in ids:
            yield CoreDataObject(
                type_=object_type,
                id_=id_,
                attributes=params
            )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import importlib
import typing
from datetime import datetime
from typing import Any

from flask import Blueprint, request

from .auth import ForbiddenError
from .misc import (
    CtxGetter,
    default_ctx_getter
)
from ..core import (
    DataObject,
    DataSourceError,
    DataSourceFilter
)

if typing.TYPE_CHECKING:
    from ..prefect import PrefectDataSource
    from ..sql import SqlDataSource


def action_blueprint(
    sql_ds: SqlDataSource,
    prefect_ds: PrefectDataSource,
    role: str | None = 'exporter',
    url_prefix: str = '/run-action',

    ctx_getter: CtxGetter = default_ctx_getter,
) -> Blueprint:
    """
    A flask `Blueprint` providing endpoints
    for applying actions on rows of a table.

    E.g. sending samples to a flow run using
    `PrefectDataSource`.
    """

    bp = Blueprint(
        'actions',
        __name__,
        url_prefix=url_prefix
    )

    def __check_required_fields(
        body: dict[str, Any],
        required_fields: list[str] = ['ids', 'action_name', 'object_type']
    ) -> None:

        if not all(field in body for field in required_fields):
            raise DataSourceError(
                'Bad Request',
                f'You must specify all of: "{", ".join(required_fields)}"',
                400
            )

    def __get_action(
        object_type: str,
        action_name: str
    ) -> DataObject:

        f = DataSourceFilter(
            and_={
                'name': {
                    'eq': {
                        'value': action_name
                    }
                },
                'object_type': {
                    'eq': {
                        'value': object_type
                    }
                }
            }
        )

        action_list = list(
            sql_ds.get_list(
                'action',
                object_filters=f
            )
        )

        if not action_list:
            raise DataSourceError(
                'Not Found',
                'The specified action was not found',
                404
            )

        return action_list[0]

    def __insert_flow_run(
        action: DataObject,
        flow_params: dict[str, Any],
        user_id: str
    ) -> tuple[str, str]:

        flow_name = action.flow_name
        flow_run = prefect_ds.data_object_factory(
            'flow_run',
            attributes={
                'flow_name': flow_name,
                'deployment_name': flow_name,
                'parameters': flow_params,
                'tags': [
                    'app_name:portal',
                    f'user_id:{user_id}'
                ],
            }
        )

        inserted_run_data = list(
            prefect_ds.insert(
                'flow_run',
                [flow_run]
            )
        )[0]

        return inserted_run_data.id, inserted_run_data.name

    @bp.post('')
    def action():

        ctx = ctx_getter()
        user_id = ctx.user_id

        if role is not None and role not in ctx.roles:
            raise ForbiddenError()

        body: dict[str, Any] = request.json.get('data', {})

        __check_required_fields(body)

        ids: list[str] = body['ids']
        action_name: str = body['action_name']
        object_type: str = body['object_type']
        params: dict[str, Any] = body.get('params', {})

        action = __get_action(object_type, action_name)

        action_params = (
            action.params
            if action.params
            else {}
        )

        if action.flow_name:
            flow_params = {
                'extra_params': {
                    **params,
                    **action_params,
                },
                'user_id': user_id,
                'object_type': object_type,
                'ids': ids
            }

            flow_run_id, flow_run_name = __insert_flow_run(
                action,
                flow_params,
                user_id
            )

            user_action_params = {
                **params,
                **action_params,
                'ids': ids,
                'flow_run_id': flow_run_id,
                'flow_run_name': flow_run_name
            }

        elif action.class_name:
            # Try to import the class from tol.actions first, then fall back to main.actions
            action_class = None
            try:
                tol_actions_module = importlib.import_module('tol.actions')
                if hasattr(tol_actions_module, action.class_name):
                    action_class = getattr(tol_actions_module, action.class_name)

                if action_class is None:
                    main_actions_module = importlib.import_module('main.actions')
                    if hasattr(main_actions_module, action.class_name):
                        action_class = getattr(main_actions_module, action.class_name)

            except ImportError:
                raise DataSourceError(
                    'Action Class Import Error',
                    'Class not found in tol.actions or main.actions',
                    500
                )

            if action_class is None:
                raise DataSourceError(
                    'Action Class Not Found',
                    f'Action class "{action.class_name}" not found in tol.actions or main.actions',
                    404
                )

            class_params = {**action_params, **params}

            action_instance = action_class()
            status = action_instance.run(ids=ids, params=class_params,
                                         object_type=object_type, datasource=sql_ds)

            user_action_params = {
                **params,
                **action_params,
                'ids': ids,
                'status': status
            }

        else:
            raise DataSourceError(
                'Invalid Action',
                'No Actions are defined',
                400
            )

        user = sql_ds.get_one('user', user_id)

        user_action = sql_ds.data_object_factory(
            'user_action',
            attributes={
                'ids': ids,
                'params': user_action_params,
                'created_at': datetime.now()
            },
            to_one={
                'user': user,
                'action': action
            }
        )
        sql_ds.insert('user_action', [user_action])

        return {'success': True}, 200

    return bp



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


import json
from io import BytesIO
from typing import Dict

from minio import Minio

from . import JsonDataSource
from ..core import (
    DataSource,
    DataSourceError,
)


class S3JsonDataSource(
    JsonDataSource
):
    """
    A subclass of JsonDataSource that loads JSON data directly from an S3 bucket.
    """

    def __init__(
        self,
        config: Dict,
        secure: bool = True
    ) -> None:
        DataSource.__init__(
            self,
            config=config,
            expected=[
                'uri',
                'type',
                'id_attribute',
                'mappings',
                's3_host',
                's3_access_key',
                's3_secret_key'
            ]
        )

        self.config = config
        self.id_attribute = config.get('id_attribute')

        bucket, object_name = self._extract_object_and_bucket(self.uri)

        # Initialize MinIO client
        self.minio_client = Minio(
            self.s3_host,
            access_key=self.s3_access_key,
            secret_key=self.s3_secret_key,
            secure=secure
        )

        # Load JSON data from S3
        raw_data = self._load_json_from_s3(bucket, object_name)

        # Set raw data
        self._raw_data = raw_data
        self._keyed_by_id = {
            v[self.id_attribute]: v
            for v in self._raw_data
            if self.id_attribute in v
        }

    def _load_json_from_s3(self, bucket: str, object_name: str):
        """Fetch and load JSON data from an S3 bucket."""
        try:
            response = self.minio_client.get_object(bucket, object_name)
            json_data = json.load(BytesIO(response.read()))  # Read and parse JSON
            return json_data
        except Exception as e:
            raise DataSourceError(f'Failed to load JSON from S3: {e}')

    def _extract_object_and_bucket(self, uri: str):
        """Extract the bucket and object name from an S3 URI."""
        if uri.startswith('s3://'):
            uri = uri[5:]
        parts = uri.split('/')
        bucket = parts[0]
        object_name = '/'.join(parts[1:])
        return bucket, object_name



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

# DO NOT WRITE ANYTHING NEW THAT USES THESE FUNCTIONS AS THEY WILL BE DELETED AT SOME POINT


from dateutil.parser import parse as dateutil_parse

from . import sts_requests


def get_setting(setting_key):
    r = sts_requests.get(f'/settings/{setting_key}')
    if r.status_code != 200:
        print(r.status_code)
        return None
    return r.json()['data']['value']


def update_setting(setting_key, new_value):
    r = sts_requests.put(
        '/settings',
        json={
            'key': setting_key,
            'value': new_value
        }
    )
    return r.status_code == 200


def get_datetime_setting(setting_key):
    value = get_setting(setting_key)
    if value is None:
        return None
    return dateutil_parse(value)


def update_datetime_setting(setting_key, new_value, fmt=None):
    string_datetime = str(new_value) if fmt is None else new_value.strftime(fmt)
    return update_setting(
        setting_key,
        string_datetime
    )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .cast import *  # noqa
from .ext import ext  # noqa
from .factory import create_sql_datasource  # noqa
from .model import model_base  # noqa
from .model import *  # noqa
from .session import create_session_factory  # noqa
from .sql_datasource import SqlDataSource  # noqa



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class TolidSpecimenToElasticTolidConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        ret = self._data_object_factory(
            'tolid',
            data_object.id,
            attributes={
                'created_at': data_object.created_at,
                'requested_taxonomy_id': data_object.requested_taxonomy_id,
                'legacy_name': data_object.legacy_name,
            },
            to_one={
                'species': self._data_object_factory(
                    'species',
                    data_object.species.id
                ),
                'specimen': self._data_object_factory(
                    'specimen',
                    data_object.specimen_id
                )
            }
        )
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Optional

from flask import request

from .auth_context import AuthContext
from ...api_client.exception import UnauthenticatedError


Authenticator = Callable[[AuthContext, Optional[str]], None]
"""
A `Callable` that should authenticate the user by:

- adding user details, such as `user_id`, to the
  given request-global `AuthContext`, given a
  `str` token (or `None`)
- raising an `UnauthenticatedError` if invalid
"""


TokenGetter = Callable[[str], Optional[str]]
MethodGetter = Callable[[], str]


def quick_and_dirty_auth(
    omnipotent_token: str,

    omnipotent_user_id: str = '666666',
    token_getter: TokenGetter = lambda h: request.headers.get(h),
    token_header: str = 'token',
    method_getter: MethodGetter = lambda: request.method,
    excluded_methods: list[str] = ['GET']
) -> Authenticator:
    """
    Compares the "token" header to a given
    `omnipotent_value`. There is only one valid token,
    and this is authorised to do everything.

    Override the `excluded_methods` kwarg with a list of
    HTTP methods on which to permit requests without
    checking the user's token. (defaults to just "GET").
    Otherwise, a token is required.

    DO NOT use this function, if it can be avoided!
    """

    def __authenticator(ctx: AuthContext) -> None:
        method = method_getter()
        if method in excluded_methods:
            return

        token = token_getter(token_header)
        if token is None:
            raise UnauthenticatedError(
                'No token was provided'
            )
        if token != omnipotent_token:
            raise UnauthenticatedError(
                'The token provided is invalid.'
            )

        ctx.user_id = omnipotent_user_id

    return __authenticator



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable

from .detail_getter import DetailGetter
from .relational import Relational
from ..datasource_filter import AndFilter, DataSourceFilter


if typing.TYPE_CHECKING:
    from ..data_object import DataObject


class Summariser(
    DetailGetter,
    Relational,
    ABC
):

    @abstractmethod
    def _summarise(
        self,
        summary_object: DataObject,
        ext_and: dict[str, Any] | None = None,
    ) -> None:
        """
        Summarises according to the given `DataObject`
        summary instance.
        """

    def summarise_all(
        self,
        summary_objects: Iterable[DataObject],
    ) -> None:
        """
        Summarises, across all types, using the given summary
        object-config instances.
        """

        for obj in summary_objects:
            self._summarise(obj)

    def summarise_type(
        self,
        summary_objects: Iterable[DataObject],
        source_object_type: str,
    ) -> None:
        """
        Summarises, for only the given `object_type`, using the given summary
        object-config instances.
        """

        filtered_summaries = self._filter_by_source_type(
            summary_objects,
            source_object_type,
        )

        for obj in filtered_summaries:
            self._summarise(obj)

    def resummarise_by_ids(
        self,
        summary_objects: Iterable[DataObject],
        source_object_type: str,
        source_object_ids: Iterable[str],
    ) -> None:
        """
        More restrictive than `summarise_all()`

        Re-summarises, using the given summary instances, only the set of changes
        affecting the `DataObject` instances of given `object_type` and
        `object_ids`.

        Returns types and IDs of what might have changed.
        """

        list_ids = list(source_object_ids)

        filtered_summaries = self._filter_by_source_type(
            summary_objects,
            source_object_type,
        )

        changed_ids: dict[str, list[str]] = {}
        for summary_obj in filtered_summaries:
            ext_and = self.get_and_filter_to_summarise(
                summary_obj,
                source_object_type,
                list_ids,
            )
            if not ext_and:
                continue
            self._summarise(
                summary_obj,
                ext_and=ext_and,
            )
            for k, v in ext_and.items():
                rel_name = k.split('.')[0]
                object_type = self.relationship_config[source_object_type].to_one[rel_name]
                if object_type not in changed_ids:
                    changed_ids[object_type] = set()
                changed_ids[object_type].update(v['in_list']['value'])
        return changed_ids

    def _mix_in_ext_and(
        self,
        object_filters: dict[str, Any] | None,
        ext_and: dict[str, Any] | None,
    ) -> DataSourceFilter:

        if not ext_and:
            return DataSourceFilter(
                and_=object_filters,
            )

        if not object_filters:
            return DataSourceFilter(
                and_=ext_and,
            )

        return DataSourceFilter(
            and_=object_filters | ext_and,
        )

    def _filter_by_source_type(
        self,
        summary_objects: Iterable[DataObject],
        source_object_type: str,
    ) -> list[DataObject]:

        return [
            s for s in summary_objects
            if s.source_object_type == source_object_type
        ]

    def get_and_filter_to_summarise(
        self,
        summary_object: DataObject,
        source_object_type: str,
        source_object_ids: Iterable[str],
    ) -> AndFilter:

        if not self.relationship_config:
            return {}

        if source_object_type not in self.relationship_config:
            return {}

        to_ones = self.relationship_config[source_object_type].to_one
        if not to_ones:
            return {}

        source_objs = list(
            self.get_by_ids(
                source_object_type,
                source_object_ids,
                requested_fields=summary_object.group_by,
            )
        )

        to_one_names = self._get_resummarised_to_one_names(
            summary_object,
            to_ones,
        )

        pairs = (
            self._ext_and_for_relationship(
                source_objs,
                relationship,
            )
            for relationship in to_one_names
        )

        return {
            k: {
                'in_list': {
                    'value': v
                }
            }
            for k, v in pairs
            if v
        }

    def _ext_and_for_relationship(
        self,
        source_objs: list[DataObject | None],
        relationship: str,
    ) -> tuple[str, list[str]]:

        relationship_id_target = f'{relationship}.id'

        relationship_ids_raw = {
            o.get_field_by_name(relationship_id_target)
            for o in source_objs
            if o is not None
        }
        relationship_ids: list[str] = [
            i for i in relationship_ids_raw
            if i is not None
        ]

        return relationship_id_target, relationship_ids

    def _get_resummarised_to_one_names(
        self,
        summary_obj: DataObject,
        to_ones: dict[str, str],
    ) -> list[str]:

        group_by: list[str] | None = summary_obj.group_by
        if not group_by:
            return list(to_ones)

        return [
            r for r in to_ones
            if any(
                g for g in group_by
                if g.startswith(r)
            )
        ]



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from .data_object import DataObject


class DataSourceError(Exception):
    """Raise to indicate that an error has occured with a DataSource."""

    def __init__(self, title: str = None, detail: str = None,
                 status_code: int = 500):
        self.title = title
        self.detail = detail
        self.status_code = status_code

    def __str__(self) -> str:
        return f'{self.title} - "{self.detail}"'


class UnknownObjectTypeException(DataSourceError):
    def __init__(self, object_type: str) -> None:
        super().__init__(
            title='Unknown Object Type',
            detail=f'The object type "{object_type}" is unknown.',
            status_code=404
        )


class NoDataObjectFactoryError(DataSourceError):
    def __init__(self, detail: str) -> None:
        super().__init__(
            'DataSource Configuration error',
            detail,
            status_code=500
        )


class NotRelationalError(DataSourceError):
    """
    Raised when trying to load exceptions on a DataObject hosted
    by a DataSource that does not implement Relational.
    """

    def __init__(self, source: DataObject) -> None:
        ds_name = type(source._host).__name__
        message = (
            f'The type "{source.type}" is hosted by a DataSource '
            f'({ds_name}) that does not support relationships.'
        )
        super().__init__(
            'DataSource Not Relational',
            message,
            status_code=500
        )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable

from dateutil.parser import parse as dateutil_parse

from ..core import DataObject

if typing.TYPE_CHECKING:
    from ..core import DataSource


DummyResource = dict[str, Any]
DummyDoc = dict[str, list[DummyResource]]


class Parser(ABC):

    def parse_iterable(
        self,
        transfers: Iterable[DummyResource]
    ) -> Iterable[DataObject]:
        """
        Parses an `Iterable` of Dummy transfer resources
        """

        return (
            self.parse(t) for t in transfers
        )

    @abstractmethod
    def parse(self, transfer: DummyResource) -> DataObject:
        """
        Parses an individual Dummy transfer resource to a
        `DataObject` instance
        """


class DefaultParser(Parser):

    def __init__(self, data_source_dict: dict[str, DataSource]) -> None:
        self.__dict = data_source_dict

    def parse(self, transfer: DummyResource) -> DataObject:
        ds = self.__dict[transfer.get('type')]
        return ds.data_object_factory(
            transfer.get('type'),
            id_=transfer.get('id'),
            attributes={
                k: (
                    dateutil_parse(v)
                    if k in ['date'] and v is not None
                    else v
                )
                for k, v in transfer.items()
                if k not in ['id', 'type', 'category']
            },
            to_one={
                'category': ds.data_object_factory(
                    'category',
                    transfer.get('category')
                ) if 'category' in transfer else None
            }
        )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Iterator, Optional

from .client import JiraClient
from .converter import (
    JiraConverter
)
from .filter import DefaultJiraFilter
from .jira_datasource import (
    JiraConverterFactory,
    JiraDataSource
)
from .parser import DefaultParser
from .sort import DefaultJiraSorter
from ..core import DataSource


class _BoldDSDict(Mapping):
    def __init__(self, api_ds: JiraDataSource) -> None:
        self.__ds = api_ds

    def __getitem__(self, __k: str) -> JiraDataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manges the instantation of:

    - `JiraConverter`
    """

    def __init__(self) -> None:
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(
        self,
        ds: DataSource
    ) -> None:

        self.__data_source = ds

    @property
    def field_mappings(self) -> Optional[dict[str, dict[str, str]]]:
        return self.__field_mappings

    @field_mappings.setter
    def field_mappings(
        self,
        fm: dict[str, dict[str, str]]
    ) -> None:

        self.__field_mappings = fm

    def jira_converter_factory(self) -> JiraConverterFactory:
        """
        Returns an instantiated `JiraConverter`.
        """

        parser = DefaultParser(
            self.__ds_dict,
            self.__field_mappings)
        return JiraConverter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _BoldDSDict(self.data_source)


def _get_client_factory(
    api_url: str,
    api_key: str
) -> Callable[[], JiraClient]:
    """
    A resonable default for creating
    a `BoldApiClient` instance
    """

    return lambda: JiraClient(
        api_url,
        api_key
    )


class _FilterFactory:
    """
    Manges the instantation of:

    - `JiraFilter`
    """

    def __init__(self) -> None:
        self.__field_mappings: Optional[dict[str, dict[str, str]]] = None

    @property
    def field_mappings(self) -> Optional[dict[str, dict[str, str]]]:
        return self.__field_mappings

    @field_mappings.setter
    def field_mappings(
        self,
        fm: dict[str, dict[str, str]]
    ) -> None:

        self.__field_mappings = fm

    def jira_filter_factory(self) -> DefaultJiraFilter:
        """
        Returns an instantiated `JiraFilter`.
        """

        return DefaultJiraFilter(self.__field_mappings)


class _SorterFactory:
    """
    Manges the instantation of:

    - `JiraSorter`
    """

    def __init__(self) -> None:
        self.__field_mappings: Optional[dict[str, dict[str, str]]] = None

    @property
    def field_mappings(self) -> Optional[dict[str, dict[str, str]]]:
        return self.__field_mappings

    @field_mappings.setter
    def field_mappings(
        self,
        fm: dict[str, dict[str, str]]
    ) -> None:

        self.__field_mappings = fm

    def jira_sorter_factory(self) -> DefaultJiraFilter:
        """
        Returns an instantiated `JiraFilter`.
        """

        return DefaultJiraSorter(self.__field_mappings)


def create_jira_datasource(
    jira_url: str,
    jira_api_key: str
) -> JiraDataSource:
    """
    Instantiates `JiraDataSource` using the given:

    - `jira_url`
    - `jira_api_key`
    """

    client_factory = _get_client_factory(
        jira_url,
        jira_api_key
    )
    manager = _ConverterFactory()
    filter_factory = _FilterFactory()
    sorter_factory = _SorterFactory()

    jira_ds = JiraDataSource(
        client_factory,
        manager.jira_converter_factory,
        filter_factory.jira_filter_factory,
        sorter_factory.jira_sorter_factory
    )

    manager.data_source = jira_ds
    manager.field_mappings = jira_ds.get_fields()
    filter_factory.field_mappings = jira_ds.get_fields()
    sorter_factory.field_mappings = jira_ds.get_fields()

    return jira_ds



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..api_client import (
    ApiDataSource,
    create_api_datasource
)
from ..core import (
    core_data_object
)


def sts(
    retries: int = 5,
    status_forcelist: list[int] | None = [429, 500, 502, 503, 504]
) -> ApiDataSource:
    sts = create_api_datasource(
        api_url=os.getenv('STS_URL', Defaults.STS_URL)
        + os.getenv('STS_API_PATH', Defaults.STS_API_PATH),
        token=os.getenv('STS_API_KEY'),
        data_prefix=os.getenv('STS_API_DATA_PATH', Defaults.STS_API_DATA_PATH),
        retries=retries,
        status_forcelist=status_forcelist,
    )
    core_data_object(sts)
    return sts



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod

from . import DataSourceFilter


class DataSourceFilterConverter(ABC):
    """
    Converts instances of `DataSourceFilter` to a
    valid filter for the remote source.
    """

    @abstractmethod
    def convert(
        self, object_type: str, object_filters: DataSourceFilter | None = None
    ) -> dict | str:
        """
        Emit a filter valid for the remote source from a `DataSourceFilter` instance
        """
        pass



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Iterable

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter


class BufferingConverter(DataObjectToDataObjectOrUpdateConverter):
    """
    A converter that buffers objects until the next object with a different ID is seen, at which
    point it yields the buffered object.
    If the next object has the same ID, it merges the attributes of the two objects together.
    """

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        pass

    __slots__ = ['__config']
    __config: Config

    def __init__(self, data_object_factory, config: Config) -> None:
        super().__init__(data_object_factory)
        self.__config = config
        self._data_object_factory = data_object_factory

    def convert_iterable(self, inputs):
        object_to_yield = None
        for data_object in inputs:
            converted_objects = list(self.convert(data_object))
            for converted_object in converted_objects:
                if object_to_yield and converted_object.id == object_to_yield.id:
                    object_to_yield = self.__merge_objects(object_to_yield, converted_object)
                else:
                    if object_to_yield:
                        yield object_to_yield
                    object_to_yield = converted_object
        if object_to_yield:
            yield object_to_yield

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        No actual coinverting is done
        """
        yield data_object

    def __merge_objects(self, obj1: DataObject, obj2: DataObject) -> DataObject:
        """
        Merges the attributes of two objects together, with the attributes of obj2 taking
        precedence over those of obj1. Deals with list attributes by concatenating them together.
        """
        if obj1.id != obj2.id:
            raise ValueError(f'Cannot merge objects with different IDs: {obj1.id} and {obj2.id}')

        merged = {}
        for key in obj1.attributes.keys() | obj2.attributes.keys():
            v1 = obj1.attributes.get(key)
            v2 = obj2.attributes.get(key)
            if isinstance(v1, list) and isinstance(v2, list):
                merged[key] = v1 + v2
            else:
                merged[key] = v2 if key in obj2.attributes else v1

        return self._data_object_factory(
            type_=obj1.type,
            id_=obj1.id,
            attributes=merged
        )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re
from typing import Dict, Generator, Iterable, List

from cachetools import LFUCache

import irods
from irods.collection import iRODSCollection
from irods.column import Criterion, In
from irods.data_object import iRODSDataObject
from irods.models import (
    Collection,
    DataObject as IDataObject,
    DataObjectMeta
)
from irods.session import iRODSSession

from ..core import (
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter
)
from ..core.operator import ListGetter


class IrodsDataSource(DataSource, ListGetter):

    def __init__(self, config: Dict):
        # uri, user, password
        super().__init__(config, expected=['host', 'port', 'user',
                                           'password', 'zone', 'query_zone',
                                           'extra_config'])
        self._initialise_irods()
        self.cache = LFUCache(100000)

    def _initialise_irods(self):
        self.irods = iRODSSession(host=self.host,
                                  port=self.port,
                                  user=self.user,
                                  password=self.password,
                                  zone=self.zone,
                                  **self.extra_config)

    def _get_collection(self, collection_id):
        c_obj = self.cache.get(collection_id)
        if c_obj is None:
            q = self.irods.query(Collection).filter(Collection.id == collection_id) \
                .add_keyword(irods.keywords.ZONE_KW, self.query_zone)
            c_id = q.one()
            c_obj = iRODSCollection(self.irods, result=c_id)
            self.cache[collection_id] = c_obj
        return c_obj

    def _format_results(self, results: Generator):
        for result in results:
            collection_id = result[IDataObject.collection_id]
            collection_object = self._get_collection(collection_id)
            data_object = iRODSDataObject(self.irods.data_objects,
                                          parent=collection_object,
                                          results=[result])
            if not re.search('(cram|bam)$', data_object.name):
                continue
            if re.search('(scraps|removed)', data_object.name):
                continue
            metadata_keys = data_object.metadata.keys()
            metadata = {}
            for key in metadata_keys:
                metadata_objects = data_object.metadata.get_all(key)
                if len(metadata_objects) == 1:
                    metadata[key] = metadata_objects[0].value
                else:
                    metadata[key] = [x.value for x in metadata_objects]
            yield {
                'data_name': data_object.name,
                'data_id': data_object.id,
                'data_create_time': data_object.create_time,
                'collection_name': collection_object.name,
                'collection_path': collection_object.path,
                'collection_create_time': collection_object.create_time,
                **metadata
            }

    def _map_keys(self, results: Generator):
        mapping = {'id_run': 'run_id',  # Illumina
                   'run': 'run_id',  # PacBio
                   'lane': 'position',  # Illumina
                   'well': 'position',
                   'type': 'file_type'}  # PacBio
        for result in results:
            # Ignore those with target = 0
            if 'id_run' in result and 'target' in result:
                if result['target'] == '0':
                    continue
            # Ignore those with a list of tags (these are legacy)
            if 'tag_index' in result and isinstance(result['tag_index'], list):
                continue
            new_obj = {}
            for k, v in result.items():
                if k in mapping:
                    new_obj[mapping[k]] = v
                else:
                    new_obj[k] = v
            yield new_obj

    def _get_run_data(self, key_name, in_list: List[str]):
        query = self.irods.query(IDataObject) \
            .add_keyword(irods.keywords.ZONE_KW, self.query_zone)

        filtered_query = query.filter(Criterion('=', DataObjectMeta.name, key_name)) \
            .filter(In(DataObjectMeta.value, in_list))
        results = filtered_query.get_results()

        return self._map_keys(self._format_results(results))

    def get_list(
        self,
        object_type: str,
        object_filters: DataSourceFilter = None,
        **kwargs
    ) -> Iterable[DataObject]:
        if object_type != 'sequencing_file':
            raise DataSourceError('Only objects of type "sequencing_file" '
                                  'are handled by IrodsDataSource')
        if object_filters is None or \
                not isinstance(object_filters.in_list, dict):
            raise DataSourceError('Filter must contain an in_list filter')

        if 'run_id' in object_filters.in_list:
            if not isinstance(object_filters.exact, dict) or \
                    'platform_type' not in object_filters.exact:
                raise DataSourceError(
                    'Filters on run_id must also contain platform_type exact filter')
            key_names = {'iseq': 'id_run',
                         'pacbio': 'run'}
            generator = self._get_run_data(
                key_names[object_filters.exact['platform_type']],
                object_filters.in_list['run_id'])
            return self._convert_dict_to_data_objects(generator)
        elif 'study_id' in object_filters.in_list:
            generator = self._get_run_data('study_id', object_filters.in_list['study_id'])
            return self._convert_dict_to_data_objects(generator)

        raise DataSourceError('Filter must contain run_id or study_id in_list filter')

    def _convert_dict_to_data_objects(self, objs: Dict) -> Iterable[DataObject]:
        return (
            self.data_object_factory('sequencing_file', attributes=obj)
            for obj in objs
        )

    @property
    def supported_types(self):
        return ['sequencing_file']



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Dict, Optional, Tuple

import requests

from .converter import EnaApiTransfer
from ..core import HttpClient


class EnaApiClient(HttpClient):
    """
    Takes ENA API transfers and connects to a remote ENA API.
    """

    def __init__(
        self,
        ena_url: str,
        ena_user: str,
        ena_password: str,
        ena_contact_name: str,
        ena_contact_email: str,
    ) -> None:
        super().__init__(
            retries=5
        )
        self.__ena_url = ena_url
        self.__ena_user = ena_user
        self.__ena_password = ena_password
        self.__ena_contact_name = ena_contact_name
        self.__ena_contact_email = ena_contact_email
        self.__type_mappings = {
            'number': 'float',
            'text': 'str',
            'array': 'List[str]',
            'date': 'datetime',
            'string': 'str',
        }

    def get_detail(
        self,
        object_type: str,
        object_ids: str,
        filter_string: Optional[str] = None
    ) -> Optional[EnaApiTransfer]:
        """
        Gets a list of ENA API transfers for the objects of a specified
        `object_type` and `object_id` or returns None if not found.
        """
        url, params = self.__detail_url(object_type, object_ids, filter_string)
        return self.__fetch_detail(url, params, text=(object_type == 'checklist'))

    def get_list(
        self,
        object_type: str,
        filter_string: Optional[str] = None
    ) -> Tuple[EnaApiTransfer, int]:
        url, params = self.__list_url(object_type, filter_string)
        return self.__fetch_list(url, params)

    def __fetch_detail(
        self,
        url: str,
        params: Dict = {},
        text: bool = False
    ) -> Optional[EnaApiTransfer]:
        """
        Fetches data from the ENA API.
        """
        session = self._get_session_with_retries()
        headers = {'Content-Type': 'application/json'}
        r = session.get(url, params=params, headers=headers)
        if r.status_code == 404:
            return []
        r.raise_for_status()
        if text:
            return r.text
        try:
            data = r.json()
            if isinstance(data, list):
                return data
            return [] if not data else [data]
        except requests.exceptions.JSONDecodeError:
            return []

    def __fetch_list(
        self,
        url: str,
        params: Dict = {}
    ) -> Optional[EnaApiTransfer]:
        """
        Fetches data from the ENA API.
        """
        session = self._get_session_with_retries()
        headers = {'Content-Type': 'application/json'}
        r = session.get(url, params=params, headers=headers)

        if r.status_code == 404:
            return []
        r.raise_for_status()
        try:
            data = r.json()
            return data if data else []
        except requests.exceptions.JSONDecodeError:
            return []

    def __detail_url(
        self,
        object_type: str,
        object_ids: str,
        filter_string: Optional[str] = None
    ) -> str:
        """
        Returns the URL and parameters for a detail query.
        """
        if object_type == 'checklist':
            ids = ','.join(str(id_) for id_ in object_ids)
            url = f'{self.__ena_url}/ena/browser/api/xml/{ids}'
            params = {}
            return url, params
        if object_type == 'submittable_taxon':
            # This is actually called separately for each taxon id
            ids = ','.join(str(id_) for id_ in object_ids)
            url = f'{self.__ena_url}/ena/taxonomy/rest/tax-id/{ids}'
            params = {}
            return url, params

        url = f'{self.__ena_url}/ena/portal/api/search'

        if object_type == 'assembly':
            query_type = 'assembly_set_accession='
            ids = 'OR assembly_set_accession='.join(f' "{id_}" ' for id_ in object_ids)
        elif object_type == 'read_run':
            query_type = 'run_accession='
            ids = 'OR run_accession='.join(f' "{id_}" ' for id_ in object_ids)
        elif object_type == 'sample':
            query_type = 'sample_accession='
            ids = 'OR sample_accession='.join(f' "{id_}" ' for id_ in object_ids)
        elif object_type == 'study':
            query_type = 'study_accession='
            ids = 'OR study_accession='.join(f' "{id_}" ' for id_ in object_ids)
        elif object_type == 'taxon':
            query_type = 'tax_id='
            ids = 'OR tax_id='.join(f' "{id_}" ' for id_ in object_ids)

        query = query_type + ids

        params = {
            'result': object_type,
            'query': query,
            'format': 'json',
            'fields': 'all'
        }

        return url, params

    def __list_url(
        self,
        object_type: str,
        filter_string: Optional[str] = None
    ) -> str:
        """
        Returns the URL and parameters for a detail query.
        """
        url = f'{self.__ena_url}/ena/portal/api/search'
        params = {
            'result': object_type,
            'query': filter_string,
            'format': 'json',
            'fields': 'all'
        }

        return url, params

    def get_fields(
        self,
        object_type
    ) -> dict:
        """
        Returns the fields for a given object type from the ENA portal API.
        """
        if object_type == 'checklist':
            return {'checklist': 'dict[str, Any]'}
        if object_type == 'submittable_taxon':
            return {
                'scientific_name': 'str',
                'formal_name': 'str',
                'rank': 'str',
                'division': 'str',
                'lineage': 'str',
                'genetic_code': 'str',
                'mitochondrial_genetic_code': 'str',
                'submittable': 'boolean',
                'binomial': 'boolean',
                'merged': 'str',
                'authority': 'str',
                'other_names': 'list[str]',
                'metagenome': 'str'
            }
        session = self._get_session_with_retries()
        r = session.get(
            self.__ena_url + '/ena/portal/api/returnFields',
            params={'result': object_type, 'format': 'json'},
            headers={
                'Content-Type': 'application/json'
            }
        )
        fields = {}
        for field in r.json():

            type_ = field['type'] if 'type' in field else 'string'
            ena_type = self.__type_mappings[type_] if type_ in self.__type_mappings else 'str'
            fields[field['columnId']] = ena_type

        return fields



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, Optional
from urllib.parse import quote

import requests

from .converter import JsonApiTransfer, JsonRelationshipConfig
from ..core import HttpClient
from ..core.datasource_error import DataSourceError
from ..core.operator import OperatorDict
from ..core.requested_fields import ReqFieldsTree


class JsonApiClient(HttpClient):
    """
    Takes JSON:API transfers and connects to a remote
    API.
    """

    def __init__(
        self,
        api_url: str,
        token: Optional[str] = None,

        data_prefix: str = '/data',
        config_prefix: str = '/_config',
        token_header: str = 'token',
        retries: int = 5,
        status_forcelist: Optional[list[int]] = None,
        merge_collections: bool | None = None,
    ) -> None:
        kwargs = {'token': token, 'token_header': token_header, 'retries': retries}
        if status_forcelist is not None:
            kwargs['status_forcelist'] = status_forcelist
        super().__init__(**kwargs)
        self.__data_url = f'{api_url}{data_prefix}'
        self.__config_url = f'{self.__data_url}{config_prefix}'
        self.__merge_collections = merge_collections

    def get_detail(
        self,
        object_type: str,
        object_id: str,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Optional[JsonApiTransfer]:
        """
        Gets a single JSON:API transfer for the object of specified
        `object_type` and `object_id`, or returns None if not found.
        """

        url = self.__detail_url(object_type, object_id)
        headers = self._merge_headers()

        params = {}
        if requested_tree:
            params['requested_fields'] = requested_tree.to_paths()

        return self.__fetch_detail(
            url,
            params=params,
            headers=headers,
        )

    def get_list_page(
        self,
        object_type: str,
        page: int,
        page_size: int,
        filter_string: Optional[str] = None,
        sort_string: Optional[str] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> JsonApiTransfer:
        """
        Gets a (paged) list-JSON:API transfer for the objects of specified
        `object_type`.
        """

        url = self.__list_url(object_type)
        params = self.__no_none_value_dict(
            page=page,
            page_size=page_size,
            filter=filter_string,
            sort_by=sort_string,
            requested_fields=requested_tree.to_paths() if requested_tree else None,
        )
        headers = self._merge_headers()
        return self.__fetch_list(
            url,
            params=params,
            headers=headers
        )

    def get_count(
        self,
        object_type: str,
        filter_string: Optional[str] = None
    ) -> JsonApiTransfer:
        """
        Gets count transfer for the objects of specified
        `object_type`.
        """

        url = self.__count_url(object_type)
        params = self.__no_none_value_dict(
            filter=filter_string
        )
        headers = self._merge_headers()
        return self.__fetch_list(url, params=params, headers=headers)

    def get_stats(
        self,
        object_type: str,
        stats_string: Optional[str],
        stats_fields_string: Optional[str],
        filter_string: Optional[str] = None
    ) -> JsonApiTransfer:
        """
        Gets stats transfer for the objects of specified
        `object_type`.
        """

        url = self.__stats_url(object_type)
        params = self.__no_none_value_dict(
            stats=stats_string,
            stats_fields=stats_fields_string,
            filter=filter_string
        )
        headers = self._merge_headers()
        return self.__fetch_list(url, params=params, headers=headers)

    def get_group_stats(
        self,
        object_type: str,
        group_by_string: str,
        stats_string: Optional[str],
        stats_fields_string: Optional[str],
        filter_string: Optional[str] = None
    ) -> JsonApiTransfer:
        """
        Gets stats transfer for the objects of specified
        `object_type`.
        """

        url = self.__group_stats_url(object_type)
        params = self.__no_none_value_dict(
            group_by=group_by_string,
            stats=stats_string,
            stats_fields=stats_fields_string,
            filter=filter_string
        )
        headers = self._merge_headers()
        return self.__fetch_list(url, params=params, headers=headers)

    def get_cursor_page(
        self,
        object_type: str,
        page_size: int,
        search_after: list[str] | None,
        filter_string: Optional[str] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> JsonApiTransfer:
        """Cursor-pagination."""

        url = self.__cursor_url(object_type)
        params = self.__no_none_value_dict(
            filter=filter_string,
            page_size=page_size,
            requested_fields=requested_tree.to_paths() if requested_tree else None,
        )
        headers = self._merge_headers()
        body = {'search_after': search_after}
        return self.__fetch_cursor(
            url,
            body,
            params=params,
            headers=headers
        )

    def delete(self, object_type: str, object_id: str) -> None:
        """
        Deletes the remote-API `DataObject` of specified type and ID.
        """
        url = self.__detail_url(object_type, object_id)
        headers = self._merge_headers()
        session = self._get_session_with_retries()
        r = session.delete(url, headers=headers)
        self.__assert_no_error(r)

    def upsert(
        self,
        object_type: str,
        transfer: JsonApiTransfer,
        merge_collections: bool | None = None,
    ) -> None:
        """
        Takes a `JsonApiTransfer` containing a `list` of
        serialized `DataObject` instances to be upserted.
        """

        if merge_collections is None:
            merge_collections = self.__merge_collections

        url = self.__upsert_url(object_type)
        params = self.__no_none_value_dict(
            merge_collections=merge_collections,
        )

        headers = self._merge_headers()
        session = self.get_session()
        r = session.post(
            url,
            headers=headers,
            params=params,
            json=transfer,
        )
        self.__assert_no_error(r)
        return r.json()

    def insert(
        self,
        object_type: str,
        transfer: JsonApiTransfer
    ) -> None:
        """
        Takes a `JsonApiTransfer` containing a `list` of
        serialized `DataObject` instances to be inserted.
        """

        url = self.__insert_url(object_type)
        headers = self._merge_headers()
        session = self.get_session()
        r = session.post(url, headers=headers, json=transfer)
        self.__assert_no_error(r)
        return r.json()

    def get_to_one_relation_recursive(
        self,
        object_type: str,
        object_id: str,
        relationship_hops: list[str]
    ) -> Optional[JsonApiTransfer]:
        """
        Fetches the nested to-one relation, on the source
        specified by the `object_type` and `object_id`,
        defined by the given `relationship_hops`.
        """

        url = self.__to_one_relation_url(
            object_type,
            object_id,
            relationship_hops
        )
        headers = self._merge_headers()
        return self.__fetch_detail(url, headers=headers)

    def get_to_many_relations_page(
        self,
        object_type: str,
        object_id: str,
        relationship_name: str,
        page: int,
        page_size: int
    ) -> JsonApiTransfer:
        """
        Fetches a page of to-many results for the given
        `relationship_name`, on the object specified by
        `object_type` and `object_id`.
        """

        url = self.__to_many_relation_url(
            object_type,
            object_id,
            relationship_name
        )
        params = {'page': page, 'page_size': page_size}
        headers = self._merge_headers()

        return self.__fetch_list(url, params=params, headers=headers)

    def config_operations(self) -> dict[str, OperatorDict]:
        """
        Fetches the supported `Operator` config for each
        `object_type` published by `api_base2`.
        """

        url = self.__config_operations_url()
        return self.__fetch_config(url)

    def config_attribute_types(self) -> dict[str, dict[str, str]]:
        """
        Fetches the `attribute_types` config for each
        `object_type` published by `api_base2`.
        """

        url = self.__config_attr_types_url()
        return self.__fetch_config(url)

    def config_attribute_metadata(self) -> dict[str, dict[str, dict[str, str | bool]]]:
        """
        Fetches the `attribute_metadata` config for each
        `object_type` published by `api_base2`.
        """

        url = self.__config_attribute_metadata_url()
        return self.__fetch_config(url)

    def config_relationships(self) -> JsonRelationshipConfig:
        """
        Fetches the `relationship_config` transfer for each
        `object_type` published by `api_base2`.
        """

        url = self.__config_rel_url()
        return self.__fetch_config(url)

    def config_write_mode(self) -> dict[str, str]:
        url = self.__config_write_mode_url()
        return self.__fetch_config(url)

    def config_return_mode(self) -> dict[str, str]:
        url = self.__config_return_mode_url()
        return self.__fetch_config(url)

    def __fetch_config(self, url: str) -> Any:
        session = self._get_session_with_retries()
        headers = self._merge_headers()
        r = session.get(url, headers=headers)
        self.__assert_no_error(r)
        return r.json()

    def __fetch_detail(
        self,
        url: str,
        params: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None
    ) -> Optional[JsonApiTransfer]:

        session = self._get_session_with_retries()
        r = session.get(url, params=params, headers=headers)
        if r.status_code == 404:
            return None
        self.__assert_no_error(r)
        return r.json()

    def __assert_no_error(
        self,
        r: requests.Response
    ) -> None:

        if r.headers.get('content-type') == 'application/json':
            return_body = r.json()

            if 'errors' in return_body:
                e: dict[str, str] = return_body['errors'][0]

                raise DataSourceError(
                    title=e.get('title'),
                    detail=e.get('detail'),
                    status_code=r.status_code
                )

        r.raise_for_status()

    def __fetch_list(
        self,
        url: str,
        params: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None
    ) -> JsonApiTransfer:

        session = self._get_session_with_retries()
        r = session.get(url, params=params, headers=headers)
        self.__assert_no_error(r)
        return r.json()

    def __fetch_cursor(
        self,
        url: str,
        body: dict[str, Any],
        params: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None
    ) -> JsonApiTransfer:

        session = self._get_session_with_retries()
        r = session.post(
            url,
            json=body,
            params=params,
            headers=headers
        )
        self.__assert_no_error(r)
        return r.json()

    def __detail_url(self, object_type: str, object_id: str) -> str:
        return f'{self.__data_url}/{object_type}/{quote(str(object_id))}'

    def __list_url(self, object_type: str) -> str:
        return f'{self.__data_url}/{object_type}'

    def __count_url(self, object_type: str) -> str:
        return f'{self.__list_url(object_type)}:count'

    def __stats_url(self, object_type: str) -> str:
        return f'{self.__list_url(object_type)}:stats'

    def __group_stats_url(self, object_type: str) -> str:
        return f'{self.__list_url(object_type)}:group-stats'

    def __cursor_url(self, object_type: str) -> str:
        return f'{self.__list_url(object_type)}:cursor'

    def __upsert_url(self, object_type: str) -> str:
        return f'{self.__list_url(object_type)}:upsert'

    def __insert_url(self, object_type: str) -> str:
        return f'{self.__list_url(object_type)}:insert'

    def __to_one_relation_url(
        self,
        object_type: str,
        object_id: str,
        relationship_hops: list[str]
    ) -> str:

        hop_string = '/'.join(relationship_hops)
        base_url = (
            f'{self.__data_url}/{object_type}:to-one/{quote(str(object_id))}'
        )
        return f'{base_url}/{hop_string}'

    def __to_many_relation_url(
        self,
        object_type: str,
        object_id: str,
        relationship_name: str
    ) -> str:
        base_url = (
            f'{self.__data_url}/{object_type}:to-many/{quote(str(object_id))}'
        )
        return f'{base_url}/{relationship_name}'

    def __config_operations_url(self) -> str:
        return f'{self.__config_url}/operations'

    def __config_attr_types_url(self) -> str:
        return f'{self.__config_url}/attribute_types'

    def __config_attribute_metadata_url(self) -> str:
        return f'{self.__config_url}/attribute_metadata'

    def __config_rel_url(self) -> str:
        return f'{self.__config_url}/relationships'

    def __config_write_mode_url(self) -> str:
        return f'{self.__config_url}/write_mode'

    def __config_return_mode_url(self) -> str:
        return f'{self.__config_url}/return_mode'

    def __no_none_value_dict(self, **kwargs) -> dict[str, Any]:
        str_params = {}
        for k, v in kwargs.items():
            if v is None:
                continue
            if isinstance(v, list) and not v:
                continue
            str_params[k] = ','.join([str(x) for x in v]) if isinstance(v, list) else str(v)
        return str_params



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional

from caseconverter import snakecase

from ..core import DataObject

if typing.TYPE_CHECKING:
    from ..core import DataSource

EnaApiResource = dict[str, Any]
EnaApiDoc = dict[str, list[EnaApiResource]]


class Parser(ABC):
    """
    Parses ENA API transfer resource `dicts` to `DataObject`
    instances.
    """

    def parse_iterable(
        self,
        transfers: Iterable[EnaApiResource]
    ) -> Iterable[DataObject]:
        """
        Parse an `Iterable` of ENA API transfer resources.
        """
        return (
            self.parse(t) for t in transfers
        )

    @abstractmethod
    def parse(
        self,
        transfer: EnaApiResource
    ) -> DataObject:
        """
        Parses an individual ENA API transfer resource to a
        `DataObject` instance.
        """


class DefaultParser(Parser):

    def __init__(
        self,
        data_source_dict: dict[str, 'DataSource']
    ) -> None:
        self.__dict = data_source_dict

    def parse(
        self,
        object_type: str,
        transfer: EnaApiResource
    ) -> DataObject:
        type_ = object_type
        ds = self.__get_data_source(type_)
        raw_attributes = transfer
        id_ = self.__get_id(type_, transfer)
        attributes = self.__convert_attributes(type_, raw_attributes)

        return ds.data_object_factory(
            type_,
            id_=id_,
            attributes=attributes,
        )

    def __get_data_source(self, type_: str) -> 'DataSource':
        if type_ not in self.__dict:
            raise ValueError(f'Data source not found for {type_}')
        return self.__dict[type_]

    def __get_id(self, type_: str, transfer: EnaApiResource) -> str:

        if type_ == 'assembly':
            return transfer['assembly_set_accession']
        elif type_ == 'read_run':
            return transfer['run_accession']
        elif type_ == 'sample':
            return transfer['sample_accession']
        elif type_ == 'study':
            return transfer['study_accession']
        elif type_ == 'taxon':
            return transfer['tax_id']
        elif type_ == 'checklist':
            return transfer['checklist_id']
        elif type_ == 'submittable_taxon':
            return transfer['taxId']

    def __convert_attributes(
        self,
        type_: str,
        attributes: Optional[dict[str, Any]]
    ) -> dict[str, Any]:
        ret = {}
        if attributes is None:
            return ret

        for k, v in attributes.items():
            if k not in ['key'] and snakecase(k) in self.__dict[type_].attribute_types[type_]:
                ret[snakecase(k)] = self.__convert_value(type_, snakecase(k), v)
        return ret

    def __convert_value(self, type_: str, attribute_name: str, value: Any) -> Any:
        attribute_type = self.__dict[type_].attribute_types[type_][attribute_name]
        if attribute_type == 'boolean':
            if isinstance(value, str):
                return value.lower() == 'true'
            return bool(value)
        return value



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from flask import Blueprint, Flask, request

from .abc import AuthManager
from .error import AuthError
from ..misc import CtxGetter, default_ctx_getter


class AuthBlueprint(Blueprint):
    """
    Holds a reference to its `AuthManager` instance.
    """

    def __init__(
        self,
        auth_manager: AuthManager,
        url_prefix: str,
        ctx_getter: CtxGetter = default_ctx_getter
    ) -> None:

        self.__manager = auth_manager
        self.__ctx_getter = ctx_getter

        super().__init__(
            'auth',
            __name__,
            url_prefix=url_prefix
        )

        self.__register_auth_endpoints()

    def register_authenticator(
        self,
        app: Flask,
        header_name: str = 'token'
    ) -> None:

        self.__manager.register(
            app,
            self.__ctx_getter,
            header_name=header_name
        )

    def __register_auth_endpoints(self) -> None:

        @self.get('/login')
        def login():
            return self.__manager.login(), 200

        @self.post('/token')
        def token():
            body: dict[str, str] = request.json

            res = self.__manager.get_token_from_callback(
                body['state'],
                body['code']
            )

            return res, 200

        @self.post('/profile')
        def profile():
            body: dict[str, str] = request.json

            res = self.__manager.create_user_profile(
                body['token']
            )

            return res, 200

        @self.post('/logout')
        @self.delete('/token')
        def logout():
            body: dict[str, str] = request.json

            self.__manager.revoke_token(body['token'])

            return {'success': True}, 200

        @self.get('/roles')
        def roles():
            ctx = self.__ctx_getter()

            return {
                'id': ctx.user_id,
                'roles': ctx.roles
            }, 200

        @self.errorhandler(AuthError)
        def auth_error(e: AuthError):
            return {
                'errors': e.errors
            }, e.status_code



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from enum import Enum


class OperatorMethod(str, Enum):
    DETAIL = 'detail GET'
    PAGE = 'list GET'
    EXPORT = 'list POST'
    COUNT = 'count GET'
    CURSOR = 'cursor GET'
    STATS = 'stats GET'
    GROUP_STATS = 'group-stat GET'
    DELETE = 'detail DELETE'
    INSERT = 'inserts POST'
    UPDATE = 'update PATCH'
    UPSERT = 'upserts POST'
    AGGREGATE = 'aggregations POST'
    TO_ONE = 'recursive to-one relation GET'
    TO_MANY = 'to-many relations GET'


class ReturnMode(str, Enum):
    """
    Describes what is returned by an instance
    of `DataSource` on write methods.
    """

    NONE = 'none'
    """
    All write methods return `None`.
    """

    PARTIAL = 'partial'
    """
    Write methods return instances of
    `DataObject`, but that may lack
    certain attributes (e.g. those
    that were not modified).
    """

    POPULATED = 'populated'
    """
    Write methods return instances of
    `DataObject` with all attributes
    populated.
    """


class RelationWriteMode(str, Enum):
    """
    Defines how a `Relational` instance
    expects relation objects during
    a write operation.
    """

    NOT_APPLICABLE = 'not applicable'
    """
    The `DataSource` does not implement
    `Relational`, or this is N/A in some
    other capacity.
    """

    FUSED = 'fused'
    """
    Relation objects should be specified
    as `to_one` without previously writing
    separately
    """

    SEPARATE = 'separate'
    """
    Relation objects must have been written
    separately before specifying in `to_one`
    """



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json

from ...api_client.exception import BadQueryArgError
from ...core import DataSourceFilter


class FilterUtils:
    @classmethod
    def parse_to_datasource_filter(cls, __key: str, __value: str) -> DataSourceFilter:
        try:
            # When making POST request, flask converts incoming JSON to a dict already
            # Only need to json.loads when using GET request
            filter_dict = __value
            if not isinstance(__value, dict):
                filter_dict = json.loads(__value)

            if isinstance(filter_dict, dict):
                dsf = DataSourceFilter(**filter_dict)
                return dsf
            raise BadQueryArgError(
                __key,
                __value,
                message=f'The {__key} must be valid JSON'
            )
        except json.JSONDecodeError:
            raise BadQueryArgError(
                __key,
                __value,
                message=f'The {__key} must be valid JSON'
            )



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from datetime import date, datetime

from tol.core import Validator
from tol.core.data_object import DataObject


class DateSortingValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances.
    For each data object (sample) check the collection date
    is not preceding the plating date
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        dates: list[str]

    __slots__ = ['__config']
    __config: Config

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()
        self.__config = config

    def _validate_data_object(self, obj: DataObject) -> None:
        # This function is used to check if the dates obtained
        # are in the standard format and the date of collection
        # is not preceding the date of plating

        previous_date = None

        for date_field in self.__config.dates:
            date_value = obj.get_field_by_name(date_field)

            # Validate that the value is a date or datetime object
            if not isinstance(date_value, (date, datetime)):
                self.add_error(
                    object_id=obj.id,
                    detail=f'{date_field} of {date_value} is not in the right date format',
                    field=self.__config.dates,
                )
                return

            # Check if dates are in ascending order
            if previous_date is not None and date_value < previous_date:
                self.add_error(
                    object_id=obj.id,
                    detail=f'Date {date_field} ({date_value})'
                    f'is before previous date ({previous_date})',
                    field=self.__config.dates,
                )
                return

            previous_date = date_value



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cache
from typing import Callable, Iterable, Optional

from .converter import (
    S3Converter
)
from ..core import (
    DataObject,
    DataSource,
)
from ..core.operator import (
    ListGetter
)
from ..services.s3_client import S3Client

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession
    from ..core.datasource_filter import DataSourceFilter

ClientFactory = Callable[[], S3Client]
S3ConverterFactory = Callable[[], S3Converter]


class S3DataSource(
    DataSource,

    # the supported operators
    ListGetter,
):
    """
    A `DataSource` that connects to a remote S3.

    Developers should likely use `create_s3_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        s3_converter_factory: S3ConverterFactory,
        bucket_name: str,
        prefix: str
    ) -> None:

        self.__client_factory = client_factory
        self.__gc_factory = s3_converter_factory
        self.bucket_name = bucket_name
        self.prefix = prefix
        super().__init__({})

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'object': {
                'bucket_name': 'str',
                'last_modified': 'datetime'
            }
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None
    ) -> Iterable[DataObject]:
        client = self.__client_factory()
        objects = client.list_objects(self.bucket_name, self.prefix)
        converted_objects = self.__gc_factory().convert_list(objects)
        return converted_objects



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .action import action_blueprint  # noqa
from .blueprint import (  # noqa
    custom_blueprint,
    data_blueprint
)
from .data_upload import data_upload_blueprint  # noqa
from .pipeline_steps import pipeline_steps_blueprint # noqa
from .system import system_blueprint  # noqa



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod

from ..core import DataSourceFilter


class EnaFilter(ABC):
    """
    Converts instances of `DataSourceFilter` to a valid ENA API query string.
    """

    @abstractmethod
    def dumps(self, filter_: DataSourceFilter) -> str:
        """
        Emit a filter string from a `DataSourceFilter` instance.
        """
        pass


class DefaultEnaFilter(EnaFilter):

    def __init__(
        self,
    ) -> None:
        super().__init__()

    __op_mappings = {
        'eq': '=',
        'lt': '<',
        'lte': '<=',
        'gt': '>',
        'gte': '>=',
        'contains': ['=', '!='],
        'in_list': ['=', '!='],
        'exists': ['is not null', 'is null']
    }

    def dumps(self, filter_: DataSourceFilter) -> str:
        filter_conditions = []
        for field, filter_list in filter_.and_.items():
            standard_field = True
            if field == 'tax_eq':
                standard_field = False
            if standard_field is True:
                for op, ops in self.__op_mappings.items():
                    if op in filter_list:
                        specific_filter = filter_list[op]
                        if 'value' in specific_filter:
                            operator = ops[0]
                            if 'negate' in specific_filter:
                                operator = ops[1]
                            if op == 'in_list':
                                filter_condition = ''
                                for v in specific_filter['value']:
                                    filter_condition += f'"{v}" OR {field}='
                                filter_condition = filter_condition[:-len(field)]
                                filter_condition = filter_condition[:-4]
                                filter_condition = filter_condition.rstrip()
                            elif op == 'contains':
                                filter_condition = str(specific_filter['value']) + '*' \
                                    if 'value' in specific_filter else ''
                            else:
                                filter_condition = str(specific_filter['value']) \
                                    if 'value' in specific_filter else ''
                            filter_conditions.append(f'{field}{operator}'
                                                     f'{filter_condition}')
            else:
                for op, ops in self.__op_mappings.items():
                    if op in filter_list:
                        specific_filter = filter_list[op]
                        if op == 'in_list':
                            values = []
                            for v in specific_filter['value']:
                                values.append(f'{field}({v})')
                            filter_condition = ' OR '.join(values)
                        else:
                            val = specific_filter['value']  \
                                if 'value' in specific_filter else ''
                            filter_condition = (f'{field}({val})')
                        filter_conditions.append(filter_condition)

        filter_string = ' AND '.join(filter_conditions)
        return filter_string if filter_string is not None else ''



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing

from sqlalchemy.orm import Session as SqlaSession

from ..core.session import DataSourceSession

if typing.TYPE_CHECKING:
    from .sql_datasource import SqlDataSource


class SqlDataSourceSession(DataSourceSession):

    def __init__(
        self,
        data_source: SqlDataSource
    ) -> None:

        self.__session = data_source.create_sqla_session()

        super().__init__(data_source)

    def __exit__(self, type_, value, tb) -> None:
        self.__session.close()

        super().__exit__(type_, value, tb)

    @property
    def _sqla_session(self) -> SqlaSession:
        """The underlying `sqlalchemy.orm.Session`."""

        return self.__session



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Iterable

from .parser import ElasticApiResource, ElasticUpdateInputResource, ElasticUpsertInputResource
from ..core import Converter, DataObject, DataSourceParser


class ElasticApiConverter(Converter[ElasticApiResource, DataObject]):
    """
    Converts from Elastic API transfers to instances of `DataObject`
    """
    __slots__ = ['__parser']
    __parser: DataSourceParser[ElasticApiResource, DataObject]

    def __init__(self, parser: DataSourceParser[ElasticApiResource, DataObject]) -> None:
        self.__parser = parser

    def convert(self, input_: ElasticApiResource) -> DataObject:
        """
        Converts an `ElasticApiTransfer` containing a detail (single) result
        """
        return self.__parser.parse(input_)

    def convert_list(
        self, input_: Iterable[ElasticApiResource]
    ) -> Iterable[DataObject]:
        """
        Converts a list of `ElasticApiTransfer`s to a list of `DataObjects`. Also returns
        a count of the total results meeting.
        """
        for elastic_resource in input_:
            yield self.__parser.parse(elastic_resource)


class ElasticUpsertInputConverter(Converter[ElasticUpsertInputResource, ElasticApiResource]):
    __slots__ = ['__parser']
    __parser: DataSourceParser[ElasticUpsertInputResource, ElasticApiResource]

    def __init__(
        self,
        parser: DataSourceParser[ElasticUpsertInputResource, ElasticApiResource]
    ) -> None:
        self.__parser = parser

    def convert(
        self,
        input_: ElasticUpsertInputResource,
    ) -> ElasticApiResource:
        return self.__parser.parse(input_)


class ElasticUpdateInputConverter(Converter[ElasticUpdateInputResource, ElasticApiResource]):
    __slots__ = ['__parser']
    __parser: DataSourceParser[ElasticUpdateInputResource, ElasticApiResource]

    def __init__(
        self,
        parser: DataSourceParser[ElasticUpdateInputResource, ElasticApiResource]
    ) -> None:
        self.__parser = parser

    def convert(
        self,
        input_: ElasticUpdateInputResource,
    ) -> ElasticApiResource:
        return self.__parser.parse(input_)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from functools import wraps
from inspect import BoundArguments
from typing import Any, Callable, Optional, Protocol

from .error import ForbiddenError
from ..misc.auth_context import AuthContext, CtxGetter, default_ctx_getter
from ...core.datasource_filter import AndFilter
from ...core.operator import OperatorMethod


class AuthInspector(Protocol):
    """
    Intercepts requests to `DataSource` instances behind `data_blueprint`.

    Raises a `ForbiddenError` for insufficient permissions.

    Can also return an update to the `DataSourceFilter().and` terms. (e.g. to
    lock down access to a user's objects by their ID).
    """

    def __call__(
        self,
        object_type: str,
        method: OperatorMethod,
        bound_args: BoundArguments
    ) -> Optional[AndFilter]:
        ...


def _assert_auth(
    ctx: AuthContext,
    required_role: Optional[str]
) -> None:

    if required_role is None:
        return

    if required_role not in ctx.roles:
        raise ForbiddenError()


def require_auth(
    arg_: Optional[Callable] = None,
    *,
    role: Optional[str] = None,

    ctx_getter: CtxGetter = default_ctx_getter
) -> Callable:
    """
    A decorator that asserts the user is logged in, and has the
    the given `role` (if specified).

    Can be used with or without parentheses, but these are mandatory
    if a keyword argument is specified.

    `role` must be specified as a keyword argument, if provided.
    """

    def decorator(func: Callable) -> Callable:

        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            ctx = ctx_getter()
            _assert_auth(ctx, role)

            return func(*args, **kwargs)

        return wrapper

    if callable(arg_):
        return decorator(arg_)

    return decorator


NO_AUTH = [
    OperatorMethod.COUNT,
    OperatorMethod.DETAIL,
    OperatorMethod.EXPORT,
    OperatorMethod.PAGE,
    OperatorMethod.TO_MANY,
    OperatorMethod.TO_ONE,
    OperatorMethod.CURSOR
]


def basic_auth_inspector(
    basic_role: str = 'basic',

    ctx_getter: CtxGetter = default_ctx_getter
) -> AuthInspector:
    """
    An `AuthInspector` factory, returning an instance
    that ignores `object_type` and:

    - permits unauthenticated access to read-only methods
    - requires the `basic_role` on an authenticated user
      otherwise
    """

    def auth_inspector(
        __object_type: str,
        method: OperatorMethod,
        bound_args: BoundArguments | None,
    ) -> None:

        if method in NO_AUTH:
            return

        roles = ctx_getter().roles
        if basic_role not in roles:
            raise ForbiddenError()

    return auth_inspector



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from ..core import (
    core_data_object
)
from ..sts import (
    StsDataSource
)


def sts_legacy(**kwargs) -> StsDataSource:
    sts_legacy = StsDataSource({
        'url': os.getenv('STS_LEGACY_URL') + os.getenv('STS_LEGACY_API_PATH'),
        'key': os.getenv('STS_API_KEY')
    })
    core_data_object(sts_legacy)
    return sts_legacy



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, Dict

from tol.core import DataObject, DataSource
from tol.core import DataSourceError, DataSourceFilter
from tol.core.validate import Validator
from tol.sources.tolid import tolid


class TolidValidator(Validator):
    """
    Validates that a stream of `DataObject` instances
    contains unique Tol IDs.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        species_id_field: str
        specimen_id_field: str
        error_ignore_field: str
        error_ignore_value: str
        warning_detail: str = 'Species not found in Tol ID source'

    __slots__ = ['__config', '__datasource', '__cached_species_id', '__cached_tolids']
    __config: Config
    __datasource: DataSource
    __cached_species_ids: Dict[str, Any]
    __cached_tolids: Dict[str, Any]

    def __init__(
        self,
        config: Config,
        datasource=tolid(),
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = config
        self.__datasource = datasource
        self.__cached_species_ids = {}
        self.__cached_tolids = {}

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:
        self.__warning_on_species_not_in_tolid(obj=obj)
        self.__error_on_specimen_id_and_taxon_not_matching_tolid(obj=obj)

    def __warning_on_species_not_in_tolid(
        self,
        obj: DataObject,
    ) -> None:

        obj_species_id = obj.get_field_by_name(self.__config.species_id_field)
        if self.__config.species_id_field in obj.attributes:
            try:
                if obj_species_id not in self.__cached_species_ids:
                    self.__cached_species_ids[obj_species_id] = (
                        self.__datasource.get_one('species', obj_species_id) is not None)

            except DataSourceError as e:
                if e.status_code == 404:
                    self.__cached_species_ids[obj_species_id] = False

        species_in_tolid = self.__cached_species_ids[obj_species_id]
        if species_in_tolid is False:
            self.add_warning(
                object_id=obj.id,
                detail=self.__config.warning_detail,
                field=self.__config.species_id_field,
            )

    def __error_on_specimen_id_and_taxon_not_matching_tolid(
        self,
        obj: DataObject,
    ) -> None:

        if (obj.get_field_by_name(self.__config.error_ignore_field) is
                self.__config.error_ignore_value):
            return

        if self.__config.specimen_id_field in obj.attributes:
            specimen_id = obj.get_field_by_name(self.__config.specimen_id_field)
            if specimen_id not in self.__cached_tolids:
                f = DataSourceFilter()
                f.and_ = {'specimen_id': {'eq': {'value': specimen_id}}}
                self.__cached_tolids[specimen_id] = list(self.__datasource.get_list(
                    'specimen',
                    object_filters=f
                ))

            if (len(self.__cached_tolids[specimen_id]) == 0):
                return
            else:
                taxons = set()
                for tolid_ in self.__cached_tolids[specimen_id]:
                    taxons.add(str(tolid_.species.id))

                if str(obj.get_field_by_name(self.__config.species_id_field)) not in taxons:
                    self.add_error(
                        object_id=obj.id + 1,
                        detail=f'Specimen ID {specimen_id} does not match Taxon ID '
                               f'{obj.get_field_by_name(self.__config.species_id_field)}'
                               'in TolID source',
                        field=[self.__config.specimen_id_field, self.__config.species_id_field]
                    )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Iterable

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter


class PrefixFieldConverter(DataObjectToDataObjectOrUpdateConverter):

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field_name: str
        prefix: str

    __slots__ = ['__config']
    __config: Config

    def __init__(self, data_object_factory, config: Config) -> None:
        super().__init__(data_object_factory)
        self.__config = config
        self._data_object_factory = data_object_factory

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        Ensures the configured field value
        starts with the configured prefix. If the field is None, it is
        left as-is.
        """

        value = data_object.get_field_by_name(
            self.__config.field_name
        )

        if value is not None:
            value_str = str(value)
            if not value_str.startswith(self.__config.prefix):
                value = f'{self.__config.prefix}{value_str}'

        ret = self._data_object_factory(
            data_object.type,
            data_object.id,
            attributes={
                **data_object.attributes,
                self.__config.field_name: value
            }
        )
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cache
from typing import Callable, Iterable, List, Optional

from .client import GoatApiClient
from .converter import (
    GoatApiConverter
)
from .filter import (
    GoatFilter
)
from ..core import (
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter
)
from ..core.operator import (
    DetailGetter,
    ListGetter,
    PageGetter,
    Relational
)
from ..core.relationship import RelationshipConfig

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession

ClientFactory = Callable[[], GoatApiClient]
FilterFactory = Callable[[], GoatFilter]
GoatConverterFactory = Callable[[], GoatApiConverter]


class GoatDataSource(
    DataSource,

    # the supported operators
    DetailGetter,
    ListGetter,
    PageGetter,
    Relational
):
    """
    A `DataSource` that connects to a remote GoaT API.

    Developers should likely use `create_goat_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        goat_converter_factory: GoatConverterFactory,
        filter_factory: FilterFactory,
    ) -> None:

        self.__client_factory = client_factory
        self.__gc_factory = goat_converter_factory
        self.__filter_factory = filter_factory
        super().__init__({})

    def __get_filter_string(
        self,
        object_filters: Optional[DataSourceFilter]
    ) -> Optional[str]:

        if object_filters is None:
            return ''
        return self.__filter_factory().dumps(object_filters)

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'taxon': {
                'scientific_name': 'str',
                'common_name': 'str',
                'synonym': 'List[str]',
                'tolid_prefix': 'str',
                'assembly_level': 'str',
                'genome_size': 'int',
                'chromosome_number': 'int',
                'haploid_number': 'int',
                'ploidy': 'int',
                'echabs92': 'str',
                'habreg_2017': 'str',
                'marhabreg-2017': 'str',
                'waca_1981': 'str',
                'isb_wildlife_act_1976': 'str',
                'protection_of_badgers_act_1992': 'str',
                'family_representative': 'List[str]',
                'lineage': 'List[str]',
                'sample_collected': 'List[str]',
                'country_list': 'List[str]',
            }
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str | int],
        **kwargs,
    ) -> Iterable[Optional[DataObject]]:
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')

        object_ids_str = [str(object_id) for object_id in object_ids]
        client = self.__client_factory()
        goat_response, _ = client.get_detail(object_type, object_ids_str)
        goat_converter = self.__gc_factory()

        converted_objects, _ = goat_converter.convert_list(goat_response) \
            if goat_response is not None else ([], 0)
        return self.sort_by_id(converted_objects, object_ids_str)

    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None,
        session: Optional[OperableSession] = None
    ) -> tuple[Iterable[DataObject], int]:
        if page_size is None:
            page_size = self.get_page_size()
        filter_string = self.__get_filter_string(object_filters)
        objects, total = self.__client_factory().get_list_page(
            object_type,
            page=page_number,
            page_size=page_size,
            filter_string=filter_string,
            sort_by=sort_by
        )
        converted_objects, _ = self.__gc_factory().convert_list(objects)
        return converted_objects, total

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None
    ) -> Iterable[DataObject]:
        # There is no way to page beyond 10000 in GoaT. After discussions with GoaT they
        # users to set off long queries. Therefore, we just set a large size and no offset here.
        filter_string = self.__get_filter_string(object_filters)
        objects, total = self.__client_factory().get_list_page(
            object_type,
            page_size=10000000,  # large enough not to matter
            filter_string=filter_string,
        )
        converted_objects, _ = self.__gc_factory().convert_list(objects)
        return iter(converted_objects)

    @property
    @cache
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        rc_taxon = RelationshipConfig()
        rc_taxon.to_one = {
            rank: 'taxon'
            for rank in self.get_ranks()
        }
        return {
            'taxon': rc_taxon
        }

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Optional[DataObject]:
        # If we are here then the relationship has not been initialised
        return None

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Iterable[DataObject]:

        return []

    def get_ranks(self) -> List[str]:
        return ['species', 'genus', 'family', 'superfamily', 'order', 'class',
                'phylum', 'kingdom', 'superkingdom', 'domain']



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from ..core import (
    core_data_object
)
from ..elastic import (
    ElasticDataSource,
    create_elastic_datasource
)


def elastic(
        environment: str = None,
        product: str = None,
        **kwargs
) -> ElasticDataSource:

    # Set up the correct environment. Can be passed in as a parameter
    # or be ELASTIC_ENVIRONMENT environment variable
    # or not be set
    if environment is None:
        environment = os.getenv('ELASTIC_ENVIRONMENT', 'production')
    if product is None:
        product = os.getenv('ELASTIC_PRODUCT', 'portal')
    index_suffix = f'-{product}' if product else ''
    index_suffix += f'-{environment}' if environment else ''
    elastic = create_elastic_datasource({
        'uri': os.getenv('ELASTIC_URI'),
        'user': os.getenv('ELASTIC_USER'),
        'password': os.getenv('ELASTIC_PASSWORD'),
        'index_prefix': os.getenv('ELASTIC_INDEX_PREFIX', '') + index_suffix},
        **kwargs
    )
    core_data_object(elastic)
    return elastic



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import itertools
import typing
from abc import ABC
from typing import Any, Iterable, Optional

import more_itertools

from ._writer import _Writer

if typing.TYPE_CHECKING:
    from ..data_object import DataObject, ErrorObject
    from ..session import OperableSession


class Inserter(_Writer, ABC):
    """
    Inserts new `DataObject` instances into a `DataSource`.

    Fails if they are already present.
    """

    def insert(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[OperableSession] = None,
        **kwargs: Any,
    ) -> Iterable[DataObject | ErrorObject] | None:
        """
        Inserts the given `DataObject` instances
        of specified type.
        """
        inserted = ()
        for batch in more_itertools.chunked(objects, self.write_batch_size):
            inserted = itertools.chain(inserted, self.insert_batch(
                object_type,
                batch,
                session=session,
                **kwargs
            ))
        return inserted

    def insert_batch(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[OperableSession] = None,
        **kwargs: Any,
    ) -> Iterable[DataObject | ErrorObject] | None:
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import List

from tol.core import DataObject, DataSource
from tol.core.validate import Validator
from tol.sources.sts import sts


class StsFieldsValidator(Validator):
    """
    Validates that a stream of `DataObject` instances
    contains fields that observe the validations in STS
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        project_code: str

    __slots__ = ['__config', '__datasource', '__fields']
    __config: Config
    __datasource: DataSource
    __fields: List[str | int | float]

    def __init__(
        self,
        config: Config,
        datasource: DataSource = sts(),  # For testing
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = config
        self.__datasource = datasource
        self.__fields = self.__initialize_fields_from_datasource()

    def __initialize_fields_from_datasource(self) -> List[str | int | float]:
        return {
            field.get('data_input_key'): field
            for field in self.__datasource.get_one(
                'project', self.__config.project_code
            ).template.get('data_fields', [])
            if field.get('in_manifest')
        }

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:
        for field in self.__fields.values():
            # Ignore inactive fields
            if field.get('status') == 'Inactive':
                continue
            # Get the value from the data object
            field_value = obj.get_field_by_name(field.get('data_input_key'))
            if isinstance(field_value, list):
                field_value = ' | '.join(str(v) for v in field_value)

            # mandatory_input fields must be present
            if field.get('mandatory_input') and field.get('data_input_key') not in obj.attributes:
                self.add_error(
                    object_id=obj.id,
                    detail=f'Field {field.get("data_input_key")} is required '
                           f'for project',
                    field=field.get('data_input_key'),
                )
                continue

            # Skip further validations if validation is not mandatory
            if not field.get('mandatory_validation'):
                continue

            # Mandatory validation fields must have a value
            if field_value is None or field_value == '':
                self.add_error(
                    object_id=obj.id,
                    detail=f'Field {field.get("data_input_key")} is required to have a value '
                           f'for project.',
                    field=field.get('data_input_key'),
                )
                continue

            # Allowed values
            if field.get('allowed_values'):
                allowed_values = [
                    value.get('value') for value in field.get('allowed_values', [])
                ]
                if field_value not in allowed_values:
                    self.add_error(
                        object_id=obj.id,
                        detail=f'Field {field.get("data_input_key")} value '
                               f'"{field_value}" not found in allowed values '
                               f'{allowed_values} for project.',
                        field=field.get('data_input_key'),
                    )

            if field.get('type') in ['String', 'TextArea']:
                self.__validate_string(obj, field, field_value)

            if field.get('type') in ['Integer', 'Decimal', 'Percentage']:
                self.__validate_number(obj, field, field_value)

            if field.get('type') in ['Boolean']:
                self.__validate_boolean(obj, field, field_value)

            if field.get('type') in ['Date']:
                self.__validate_date(obj, field, field_value)

    def __validate_string(
        self,
        obj: DataObject,
        field: dict,
        field_value: str | int | float | None
    ) -> None:
        # Check type is a string
        # if not isinstance(field_value, str):
        #     self.add_error(
        #         object_id=obj.id,
        #         detail=f'Field {field.get("data_input_key")} value '
        #                 f'"{field_value}" is not a string for project.',
        #         field=field.get('data_input_key'),
        #     )
        #     return

        # Min/Max validations for string
        if field.get('min') and len(field_value) < field.get('min'):
            self.add_error(
                object_id=obj.id,
                detail=f'Field {field.get("data_input_key")} value '
                       f'"{field_value}" is shorter than minimum length '
                       f'"{field.get("min")}" for project.',
                field=field.get('data_input_key'),
            )
        if field.get('max') and len(field_value) > field.get('max'):
            self.add_error(
                object_id=obj.id,
                detail=f'Field {field.get("data_input_key")} value '
                       f'"{field_value}" is longer than maximum length '
                       f'"{field.get("max")}" for project.',
                field=field.get('data_input_key'),
            )

    def __validate_number(
        self,
        obj: DataObject,
        field: dict,
        field_value: str | int | float | None
    ) -> None:
        # Check type is a number
        if not isinstance(field_value, (int, float)):
            self.add_error(
                object_id=obj.id,
                detail=f'Field {field.get("data_input_key")} value '
                       f'"{field_value}" is not a number for project.',
                field=field.get('data_input_key'),
            )
            return

        # Min/Max validations for number
        if field.get('min') is not None and field_value < field.get('min'):
            self.add_error(
                object_id=obj.id,
                detail=f'Field {field.get("data_input_key")} value '
                       f'"{field_value}" is less than minimum value '
                       f'"{field.get("min")}" for project.',
                field=field.get('data_input_key'),
            )
        if field.get('max') is not None and field_value > field.get('max'):
            self.add_error(
                object_id=obj.id,
                detail=f'Field {field.get("data_input_key")} value '
                       f'"{field_value}" is greater than maximum value '
                       f'"{field.get("max")}" for project.',
                field=field.get('data_input_key'),
            )

    def __validate_boolean(
        self,
        obj: DataObject,
        field: dict,
        field_value: str | int | float | None
    ) -> None:
        # Check type is a boolean
        if field_value not in ['Y', 'N']:
            self.add_error(
                object_id=obj.id,
                detail=f'Field {field.get("data_input_key")} value '
                       f'"{field_value}" is not a boolean (Y/N) for project.',
                field=field.get('data_input_key'),
            )

    def __validate_date(
        self,
        obj: DataObject,
        field: dict,
        field_value: str | int | float | None
    ) -> None:
        if not isinstance(field_value, datetime):
            self.add_error(
                object_id=obj.id,
                detail=f'Field {field.get("data_input_key")} value '
                       f'"{field_value}" is not a date string for project.',
                field=field.get('data_input_key'),
            )
            return
        if field.get('range_limit'):
            earliest_date = datetime.now() - timedelta(days=field.get('min'))
            latest_date = datetime.now() + timedelta(days=field.get('max'))
            if field_value < earliest_date or field_value > latest_date:
                self.add_error(
                    object_id=obj.id,
                    detail=f'Field {field.get("data_input_key")} value '
                           f'"{field_value}" is not within the allowed date '
                           f'range for project.',
                    field=field.get('data_input_key'),
                )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class BioscanImageToElasticSampleUpdateConverter(
    DataObjectToDataObjectOrUpdateConverter
):
    """
    Converter for Bioscan images stored in S3 Buckets
    into an Updater Data Object for Elastic (Portal's data source)
    """
    def convert(self, input_: DataObject) -> Iterable[DataObjectUpdate]:
        """
        Converts a Data Object of type 'object' representing a Bioscan image
        stored in an S3 bucket into an Updater Data Object for Elastic (Portal's data source)

        :param input_: The input data object from an S3 bucket storing a Bioscan image.
        Has type 'object' and must have an id (its file name on S3)

        :returns: A generator of output `DataObjectUpdate`s which can be used to update
        the the sample data objects in Elastic
        """
        # Ensure a data object has been passed in
        if input_ is None:
            return

        # Ensure the data object has an id
        if input_.id is None:
            return

        # Get S3 bucket name from input Bioscan image data object
        bucket_name = input_.attributes['bucket_name']

        # The id of a Bioscan image is its S3 file name
        bioscan_image_file_name = input_.id

        # This file name contains the sample id of the Bioscan image
        # The following function extracts this
        sample_id = self.__extract_sample_id_from_s3_file_name(bioscan_image_file_name)

        # Construct the URL for the Bioscan image using this information
        bioscan_image_url = f'https://{bucket_name}.cog.sanger.ac.uk/{bioscan_image_file_name}'

        attributes = {
            # For a Bioscan image, the specimen id and sample id are the same,
            # because a Bioscan sample is the whole specimen (an entire insect)
            'sts_specimen.id': sample_id,
            'url': bioscan_image_url,
            'modified': input_.attributes['last_modified']
        }

        yield (None, attributes)  # type: ignore (Linter does not properly recognise type here)

    def __extract_sample_id_from_s3_file_name(self, file_name: str) -> str:
        """
        Extracts the sample id of a Bioscan image from its S3 file name,
        which is assumed to be in the format

        '`<prefix>`/`<sample id>`.`<suffix>`',
        where `<prefix>/` is optional

        :returns: sample_id
        """
        # Remove file prefix (even if not present)
        file_name_without_prefix = file_name.split('/')[-1]

        # Remove file suffix
        sample_id = file_name_without_prefix.split('.')[0]

        return sample_id



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, List, Optional, Union

from .parser import Parser
from ..core import DataObject

EnaApiObject = dict[str, Any]
EnaApiTransfer = List[dict[
    str,
    Union[EnaApiObject, list[EnaApiObject]]
]]


class EnaApiConverter():
    """
    Converts from ENA API transfers to instances of
    `DataObject`.
    """

    def __init__(
        self,
        parser: Parser
    ) -> None:

        self.__parser = parser

    def convert(self, input_: EnaApiTransfer) -> DataObject:
        """
        Convert an EnaApiTransfer containing a detail (single) result.
        """
        return self.__parser.parse(input_)

    def convert_list(
        self,
        object_type: str,
        input_: EnaApiTransfer
    ) -> tuple[list[DataObject], Optional[int]]:
        """
        Converts an EnaApiTransfer containing a list of results. Also
        returns a count of the total number of results.
        """
        return [
            self.__parser.parse(object_type, json_obj)
            for json_obj in input_
        ], len(input_)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional


class AttributeMetadata(ABC):
    @abstractmethod
    def get_display_name(
            self,
            object_type: str,
            attribute_name: str) -> str:
        """
        Gets the attribute display name
        """

    @abstractmethod
    def is_available_on_relationships(
            self,
            object_type: str,
            attribute_name: str) -> bool:
        """
        Is the attribute available on a to_one related object?
        This is not always the case, for example in elastic, only
        enriched attributes will be available.
        """

    @abstractmethod
    def get_cardinality(
            self,
            object_type: str,
            attribute_name: str) -> Optional[int]:
        """
        The approximate cardinality of the attribute, i.e. how many different
        values there are. This can be used for making decisions in the UI
        about how to display or filter.
        """

    @abstractmethod
    def get_description(
            self,
            object_type: str,
            attribute_name: str) -> Optional[str]:
        """
        The description of an attribute. This would normally be fetched from
        and underlying metadata store.
        """

    @abstractmethod
    def get_source(
            self,
            object_type: str,
            attribute_name: str) -> Optional[str]:
        """
        The source of the attribute i.e becnhling, sts etc
        """

    @abstractmethod
    def is_authoritative(
        self,
        object_type: str,
        attribute_name: str
    ) -> bool:
        pass

    @abstractmethod
    def get_acts_as(
        self,
        object_type: str,
        attribute_name: str
    ) -> Optional[str]:
        """
        A semantic role for the attribute (e.g. 'status'), used by consumers
        to apply special behaviour.
        """


class DefaultAttributeMetadata(AttributeMetadata):
    def get_display_name(
            self,
            object_type: str,
            attribute_name: str) -> str:
        parts = attribute_name.split('_')
        words = []
        for part in parts:
            words.append(self.__normalise_word(part))
        return ' '.join(words)

    def __normalise_word(self, word: str) -> str:
        replacements = {
            'id': 'ID',
            'uid': 'UID',
            'sts': 'STS',
            'tolqc': 'ToLQC',
            'tolid': 'ToLID',
            'tol': 'ToL',
            'eln': 'ELN',
            'dna': 'DNA',
            'rna': 'RNA',
            'mlwh': 'MLWH'
        }
        if word in replacements:
            return replacements[word]
        return word.capitalize()

    def is_available_on_relationships(
            self,
            object_type: str,
            attribute_name: str) -> bool:
        """
        Is the attribute available on a to_one related object?
        This is not always the case, for example in elastic, only
        enriched attributes will be available.
        """
        return True

    def is_authoritative(
            self,
            object_type: str,
            attribute_name: str) -> bool:
        """
        Is the attribute authoritative (i.e. there are other options for this
        piece of data but this one is the most trusted)
        """
        return True

    def get_cardinality(
            self,
            object_type: str,
            attribute_name: str) -> Optional[int]:
        """
        The approximate cardinality of the attribute, i.e. how many different
        values there are. This can be used for making decisions in the UI
        about how to display or filter.
        """
        return None

    def get_description(
            self,
            object_type: str,
            attribute_name: str) -> Optional[str]:
        """
        The description of an attribute. This would normally be fetched from
        and underlying metadata store.
        """
        return None

    def get_source(
            self,
            object_type: str,
            attribute_name: str) -> Optional[str]:
        """
        The source of the attribute i.e becnhling, sts etc
        """
        return None

    def get_acts_as(
            self,
            object_type: str,
            attribute_name: str) -> Optional[str]:
        """
        A semantic role for the attribute (e.g. 'status').
        """
        return None



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional

from ._writer import _Writer

if typing.TYPE_CHECKING:
    from ..data_object import DataObject, ErrorObject
    from ..session import OperableSession


DataObjectUpdate = tuple[str, dict[str, Any]]


class Updater(_Writer, ABC):
    """
    Updates existing DataObject instances.
    """

    @abstractmethod
    def update(
        self,
        object_type: str,
        updates: Iterable[DataObjectUpdate],
        session: Optional[OperableSession] = None,
        **kwargs: Any,
    ) -> Iterable[DataObject | ErrorObject] | None:
        """
        Takes a type and an `Iterable` of ID-`DataObjectUpdate` pairs.

        For each pair, the existing `DataObject` identified by the ID is
        updated using the `DataObjectUpdate` dict of new values. Values
        that aren't mentioned remain the same as previously.
        """



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Any, Iterable

from dateutil import parser

import pandas as pd

from ..core import (
    DataObject,
    DataSource,
    DataSourceError,
    core_data_object,
)
from ..core.operator import ListGetter


class PandasDataSource(
    DataSource,
    ListGetter,
    ABC
):
    """
    Emits `DataObject` instances from a tabular
    spreadsheet (e.g. `.xlsx`).
    """

    def __init__(
        self,
        filepath: str | Path,
        sheet_name: str,
        *,
        object_type: str = 'sheet_row',
        engine: str = 'openpyxl',
        type_mapping: dict[str, str] | None = None,
    ) -> None:

        super().__init__({})

        self._object_type = object_type

        self._mappings = self.__get_mappings(
            type_mapping,
        )

        self._df = self._get_dataframe(
            filepath,
            sheet_name,
            engine,
        )

        core_data_object(self)

    @property
    def supported_types(self) -> list[str]:
        return [self._object_type]

    def get_list(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Iterable[DataObject]:

        return (
            self.__marshal_row(row_index + 2, row)  # Add 1 for header, 1 for 1-based ID
            for row_index, row
            in self._df.iterrows()
        )

    def __marshal_row(
        self,
        row_index: int,
        row: pd.Series,
    ) -> DataObject:

        attributes = self.__format_attributes(row)

        return self.data_object_factory(
            self._object_type,
            id_=str(row_index),
            attributes=attributes,
        )

    def __format_attributes(
        self,
        row: pd.Series,
    ) -> dict[str, Any]:

        return {
            k: self.__format_attribute(k, v)
            for k, v in row.items()
        }

    def __format_attribute(
        self,
        __k: str,
        __v: Any,
    ) -> Any:

        # Convert pandas Timestamp to Python datetime
        if isinstance(__v, pd.Timestamp):
            __v = datetime.fromtimestamp(__v.timestamp())

        # If float and is whole number, convert to int
        if isinstance(__v, float) and __v.is_integer():
            __v = int(__v)

        if __k not in self._mappings:
            return __v

        type_ = self._mappings[__k]

        if type_ == datetime and not isinstance(__v, datetime):
            return parser.parse(__v)

        return type_(__v)

    def __get_mappings(
        self,
        type_mappings: dict[str, str] | None,
    ) -> dict[str, type]:

        if not type_mappings:
            return {}

        return {
            k: self.__get_mapping(v)
            for k, v in type_mappings.items()
        }

    def __get_mapping(self, __v: str) -> type:
        if __v == 'int':
            return int
        elif __v == 'str':
            return str
        elif __v == 'float':
            return float
        elif __v == 'datetime':
            return datetime
        elif __v == 'bool':
            return bool

        raise DataSourceError(title='Bad Mapping Value')

    @abstractmethod
    def _get_dataframe(
        self,
        filepath: str | Path,
        sheet_name: str,
        engine: str,
    ) -> pd.DataFrame:
        raise NotImplementedError



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class ElasticSampleToBenchlingTissueUpdateConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def _convert_one(self, data_object: DataObject) -> DataObjectUpdate:
        if data_object is None:
            return
        species = data_object.to_one_relationships['sts_species']
        specimen = data_object.to_one_relationships['sts_specimen']

        if species is not None and specimen is not None:
            ret = (
                data_object.benchling_eln_tissue_id,
                {
                    'rack_id': data_object.sts_rackid,
                    'tube_well_id': data_object.sts_tubeid,
                    'tube_position': data_object.sts_pos_in_rack,
                    'scientific_name': species.sts_scientific_name,
                    'taxon_id': species.id,
                    'taxon_group_phyla':
                        species.sts_taxon_group
                        if species.sts_taxon_group else 'NA',
                    'genome_size': str(species.sts_genome_size),
                    # 'freezer': None,
                    'location': data_object.sts_labwhere_parentage,
                    'tray': data_object.sts_labwhere_name,
                    'specimen_id': specimen.id,
                    'programme_id': data_object.sts_tolid.id,
                    'biosample_id': data_object.sts_biosample_accession,
                    'biospecimen_id': data_object.calc_biospecimen_id,  # Needs work
                    'organism_part':
                        ', '.join(data_object.sts_organism_part)
                        if data_object.sts_organism_part is not None else None,
                    'lifestage': data_object.sts_lifestage,
                    'sex': data_object.sts_sex,
                    'preservation_approach': data_object.sts_preservation_approach,
                    'size_of_tissue_in_tube': data_object.sts_tissue_size,
                    'hazard_group': data_object.sts_hazard_group,
                    'date_sample_received_at_sanger':
                        data_object.sts_receive_date.strftime('%Y-%m-%d')
                        if data_object.sts_receive_date is not None else '1970-01-01',
                    'date_assigned_to_lab':
                        data_object.sts_tollab_assign_date.strftime('%Y-%m-%d')
                        if data_object.sts_tollab_assign_date is not None else '1970-01-01',
                    # 'assigned_by': ,
                    # 'lab_work_category': data_object.sts_lab_work_category,
                    'family_representative': ', '.join(species.goat_family_representative)
                        if species.goat_family_representative is not None else None,
                    'sample_set_id': data_object.sts_sampleset.id,
                    'rd_sample': data_object.sts_send_rd,
                    'sts_id': int(data_object.id),
                    # 'remaining_weight':,
                    'priority': data_object.sts_priority,
                    'project': ', '.join(data_object.sts_project),
                    'study_id': data_object.sts_sequencescape_study_id,
                    'cost_code': data_object.sts_cost_code,
                }
            )
            return ret

    def convert(self, data_object: DataObject) -> Iterable[DataObjectUpdate]:
        if data_object.benchling_eln_tissue_id is not None:
            ret = self._convert_one(data_object)
            if ret is not None:
                yield ret



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import List

from tol.core import DataObject
from tol.core.validate import Validator


class AllowedKeysValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances,
    ensuring that they only have attributes of the given
    allowed keys.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        allowed_keys: List[str]
        is_error: bool = True
        detail: str = 'Key is not allowed'

    __slots__ = ['__config']
    __config: Config

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        del kwargs
        super().__init__()
        self.__config = config

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:

        for key in obj.attributes:
            if key not in self.__config.allowed_keys:
                self.__add_result(
                    obj,
                    key,
                )

    def __add_result(
        self,
        obj: DataObject,
        key: str,
    ) -> None:

        if self.__config.is_error:
            self.add_error(
                object_id=obj.id,
                detail=self.__config.detail,
                field=key,
            )
        else:
            self.add_warning(
                object_id=obj.id,
                detail=self.__config.detail,
                field=key,
            )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from datetime import datetime
from functools import cache
from typing import Dict, Iterable, List

from dateutil.parser import parse as dateutil_parse

import gspread

from ..core import (
    DataId,
    DataObject,
    DataSource,
    DataSourceFilter
)
from ..core.operator import (
    DetailGetter,
    ListGetter,
)


class GoogleSheetDataSource(
    DataSource,
    DetailGetter,
    ListGetter,
):
    """
    mappings = {
        'object_name': {
            'worksheet_name': "Sheet name',
            'columns': {
                'id': {
                    'heading': 'Column name',
                    'type': 'int'
                }
                'field2': {
                    'heading': 'Column for field 2',
                    'type': 'int'
                }
            },
            'header_row': 2,
            'data_start_row': 4
        }
    }
    """

    def __init__(self, config: Dict):
        super().__init__(config, expected=[
            'client_secrets',
            'sheet_key',
            'mappings'])
        self._initialise_google_sheet()
        self.data = {}

    def _initialise_google_sheet(self):
        gc = gspread.service_account_from_dict(self.client_secrets)
        self.sheet = gc.open_by_key(self.sheet_key)

    @cache
    def _initialise_data(self, object_type):
        worksheet = self.sheet.worksheet(self.mappings[object_type]['worksheet_name'])
        self.data[object_type] = self._get_worksheet_vals(object_type, worksheet)

    def _get_header_mappings(self, object_type):
        ret = {}
        for attribute_name, metadata in self.mappings[object_type]['columns'].items():
            ret[metadata['heading']] = attribute_name
        return ret

    def __convert_type(self, object_type, header, value):
        header_mappings = self._get_header_mappings(object_type)
        target_type = self.mappings[object_type]['columns'][header_mappings[header]]['type']
        try:
            value = value.strip()
            if value == '':
                return None
            if target_type == 'int':
                return int(value)
            if target_type == 'float':
                return float(value)
        except ValueError:
            return None
        return value

    def _get_worksheet_vals(self, object_type, worksheet):
        vals = worksheet.get_values()
        header_row = self.mappings[object_type]['header_row']
        data_start_row = self.mappings[object_type]['data_start_row']
        headers = vals[header_row - 1]
        row_data = vals[data_start_row - 1:]
        ret = []
        for row in row_data:
            item = {}
            for i, header in enumerate(headers):
                if header in self._get_header_mappings(object_type).keys():
                    item[header] = self.__convert_type(object_type, header, row[i])
            ret.append(item)
        return ret

    def _convert_row_to_data_object(self, object_type, row):
        CoreDataObject = self.data_object_factory  # noqa N806
        attributes = {attribute_name: row[column_def['heading']]
                      for attribute_name, column_def
                      in self.mappings[object_type]['columns'].items()}
        # Sort out Booleans
        for attribute_name, attribute_value in attributes.items():
            if self.mappings[object_type]['columns'][attribute_name]['type'] == 'boolean' and \
                    attribute_value is not None:
                attributes[attribute_name] = True \
                    if attribute_value in [1, '1', 'Y', 'Yes', 'YES'] else False
        # Sort out datetimes
        for attribute_name, attribute_value in attributes.items():
            if self.mappings[object_type]['columns'][attribute_name]['type'] == 'datetime' and \
                    attribute_value is not None and not isinstance(attribute_value, datetime):
                dayfirst = self.mappings[object_type]['columns'][attribute_name].get(
                    'dayfirst', False
                )
                attributes[attribute_name] = dateutil_parse(
                    attribute_value,
                    dayfirst=dayfirst
                )
        return CoreDataObject(
            object_type,
            id_=attributes.pop('id', None),
            attributes=attributes
        )

    def _apply_filter(self, f: DataSourceFilter, rows: List, object_type: str):
        for row in rows:
            matches = True
            if f is not None:
                if f.in_list is not None:
                    for column_name, allowed_values in f.in_list.items():
                        key = self.mappings[object_type]['columns'][column_name]['heading']
                        if row[key] not in allowed_values:
                            matches = False
                if f.exact is not None:
                    for column_name, allowed_value in f.exact.items():
                        key = self.mappings[object_type]['columns'][column_name]['heading']
                        if row[key] != allowed_value:
                            matches = False
            if matches:
                yield row

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[DataId],
        **kwargs
    ) -> Iterable[DataObject]:
        self._initialise_data(object_type)
        f = DataSourceFilter()
        f.in_list = {'id': object_ids}
        rows = self._apply_filter(f, self.data[object_type], object_type)
        for row in rows:
            yield self._convert_row_to_data_object(object_type, row)

    def get_list(
        self,
        object_type: str,
        object_filters: DataSourceFilter = None,
        **kwargs
    ) -> Iterable[DataObject]:
        self._initialise_data(object_type)
        rows = self._apply_filter(object_filters, self.data[object_type], object_type)
        for row in rows:
            obj = self._convert_row_to_data_object(object_type, row)
            if obj.id is not None:
                yield obj

    @property
    @cache
    def supported_types(self):
        return list(self.mappings.keys())

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {object_type: {attribute_name: column_def['type']
                              for attribute_name, column_def in mapping['columns'].items()}
                for object_type, mapping in self.mappings.items()}



# SPDX-FileCopyrightText: 2021 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from tol.sciops.messages import CreateLabwareMessage, LabwareMessage, UpdateLabwareMessage


class MessageBuilder:
    """ Build up various message types for sending to MQ """

    def __init__(self):
        pass

    @classmethod
    def build_labware_message(cls, message: LabwareMessage):
        if isinstance(message, CreateLabwareMessage):
            return cls._build_create_labware_message(message)
        elif isinstance(message, UpdateLabwareMessage):
            return cls._build_update_labware_message(message)
        else:
            raise TypeError('Unknown message type provided')

    @staticmethod
    def _build_create_labware_message(message: CreateLabwareMessage) -> dict:
        """ Build a create_labware message """
        output = {}
        labware_dict = {}
        samples = []
        for sample in message.samples:
            sample_dict = {}
            sample_dict.update(
                {} if sample.sample_uuid is None else {'sampleUuid': sample.sample_uuid.encode()})
            sample_dict.update(
                {} if sample.study_uuid is None else {'studyUuid': sample.study_uuid.encode()})
            sample_dict.update(
                {} if sample.sanger_sample_id is None else {
                    'sangerSampleId': sample.sanger_sample_id})
            sample_dict.update(
                {} if sample.location is None else {'location': sample.location})
            sample_dict.update(
                {} if sample.supplier_sample_name is None else {
                    'supplierSampleName': sample.supplier_sample_name})
            sample_dict.update({} if sample.volume is None else {'volume': sample.volume})
            sample_dict.update(
                {} if sample.concentration is None else {'concentration': sample.concentration})
            sample_dict.update(
                {} if sample.public_name is None else {'publicName': sample.public_name})
            sample_dict.update({} if sample.taxon_id is None else {'taxonId': sample.taxon_id})
            sample_dict.update(
                {} if sample.common_name is None else {'commonName': sample.common_name})
            sample_dict.update(
                {} if sample.donor_id is None else {'donorId': sample.donor_id})
            sample_dict.update(
                {} if sample.library_type is None else {'libraryType': sample.library_type})
            sample_dict.update(
                {} if sample.country_of_origin is None else {
                    'countryOfOrigin': sample.country_of_origin})
            sample_dict.update({} if sample.sample_collection_date_utc is None else {
                'sampleCollectionDateUtc': sample.sample_collection_date_utc.timestamp() * 1000})
            sample_dict.update(
                {} if sample.cost_code is None else {'costCode': sample.cost_code})
            sample_dict.update(
                {} if sample.genome_size is None else {'genomeSize': sample.genome_size})
            sample_dict.update(
                {} if sample.accession_number is None else {
                    'accessionNumber': sample.accession_number})
            sample_dict.update(
                {} if sample.sheared_femto_fragment_size is None else {
                    'shearedFemtoFragmentSize': sample.sheared_femto_fragment_size})
            sample_dict.update(
                {} if sample.post_spri_concentration is None else {
                    'postSPRIConcentration': sample.post_spri_concentration})
            sample_dict.update(
                {} if sample.post_spri_volume is None else {
                    'postSPRIVolume': sample.post_spri_volume})
            sample_dict.update(
                {} if sample.final_nano_drop_280 is None else {
                    'finalNanoDrop280': sample.final_nano_drop_280})
            sample_dict.update(
                {} if sample.final_nano_drop_230 is None else {
                    'finalNanoDrop230': sample.final_nano_drop_230})
            sample_dict.update(
                {} if sample.final_nano_drop is None else {
                    'finalNanoDrop': sample.final_nano_drop})
            sample_dict.update(
                {} if sample.shearing_and_qc_comments is None else {
                    'shearingAndQCComments': sample.shearing_and_qc_comments})
            sample_dict.update(
                {} if sample.date_submitted_utc is None else {
                    'dateSubmittedUTC': sample.date_submitted_utc.timestamp() * 1000})
            sample_dict.update(
                {} if sample.priority_level is None else {
                    'priorityLevel': sample.priority_level})
            sample_dict.update(
                {} if sample.date_required_by is None else {
                    'dateRequiredBy': sample.date_required_by})
            sample_dict.update(
                {} if sample.reason_for_priority is None else {
                    'reasonForPriority': sample.reason_for_priority})
            samples.append(sample_dict)
        labware_dict.update(
            {} if message.labware_type is None else {'labwareType': message.labware_type})
        labware_dict.update(
            {} if message.labware_uuid is None else {'labwareUuid': message.labware_uuid.encode()})
        labware_dict.update({} if message.barcode is None else {'barcode': message.barcode})
        labware_dict['samples'] = samples
        output.update(
            {} if message.message_uuid is None else {'messageUuid': message.message_uuid.encode()})
        output.update({} if message.message_create_date_utc is None else {
            'messageCreateDateUtc': message.message_create_date_utc.timestamp() * 1000})
        output['labware'] = labware_dict
        return output

    @staticmethod
    def _build_update_labware_message(message: UpdateLabwareMessage) -> dict:
        """ Build a update_labware message """
        output = {}
        labware_updates = []
        sample_updates = []
        for upd in message.labware_updates:
            labware_updates_dict = {}
            labware_updates_dict.update(
                {} if upd.uuid is None else {'labwareUuid': upd.uuid.encode()})
            labware_updates_dict.update({} if upd.name is None else {'name': upd.name})
            labware_updates_dict.update({} if upd.value is None else {'value': upd.value})
            labware_updates.append(labware_updates_dict)
        for upd in message.sample_updates:
            sample_updates_dict = {}
            sample_updates_dict.update(
                {} if upd.uuid is None else {'sampleUuid': upd.uuid.encode()})
            sample_updates_dict.update({} if upd.name is None else {'name': upd.name})
            sample_updates_dict.update({} if upd.value is None else {'value': upd.value})
            sample_updates.append(sample_updates_dict)
        output.update(
            {} if message.message_uuid is None else {'messageUuid': message.message_uuid.encode()})
        output.update({} if message.message_create_date_utc is None else {
            'messageCreateDateUtc': message.message_create_date_utc.timestamp() * 1000})
        output['labwareUpdates'] = labware_updates
        output['sampleUpdates'] = sample_updates
        return output



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import numpy as np

import pandas as pd


def convert_csv_to_json(file_path, sep=',', usecols=None):
    csv_file = pd.read_csv(
        file_path,
        sep=sep,
        header=0,
        index_col=False,
        usecols=usecols
    )
    csv_file.replace({np.nan: None}, inplace=True)
    return csv_file.to_dict(orient='records')



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .benchling_extraction_container_to_elastic_extraction_container_converter import BenchlingExtractionContainerToElasticExtractionContainerConverter  # noqa F401
from .benchling_extraction_to_elastic_extraction_converter import BenchlingExtractionToElasticExtractionConverter  # noqa F401
from .benchling_extraction_to_elastic_sequencing_request_converter import BenchlingExtractionToElasticSequencingRequestConverter  # noqa F401
from .benchling_sequencing_request_to_elastic_sequencing_request_converter import (  # noqa F401
    BenchlingSequencingRequestToElasticSequencingRequestConverter
)
from .benchling_tissue_prep_to_elastic_tissue_prep_converter import BenchlingTissuePrepToElasticTissuePrepConverter  # noqa F401
from .benchling_entity_to_benchling_worklist_item_converter_factory import BenchlingEntityToBenchlingWorklistItemConverterFactory  # noqa F401
from .benchling_tissue_to_sts_sample_converter import BenchlingTissueToStsSampleConverter  # noqa F401
from .benchling_sample_to_elastic_sample_converter import BenchlingSampleToElasticSampleConverter  # noqa F401
from .bioscan_extra_pantheon_species_to_elastic_sample_update_converter import BioscanExtraPantheonSpeciesToElasticSampleUpdateConverter  # noqa 
from .bioscan_image_to_elastic_sample_update_converter import BioscanImageToElasticSampleUpdateConverter  # noqa F401
from .bioscan_qc_specimen_to_elastic_sample_update_converter import BioscanQcSpecimenToElasticSampleUpdateConverter  # noqa F401
from .bioscan_qc_uksi_entry_to_elastic_sample_update_converter import BioscanQcUksiEntryToElasticSampleUpdateConverter # noqa F401
from .bold_bin_to_elastic_sample_update_converter import BoldBinToElasticSampleUpdateConverter  # noqa F401
from .bold_sample_to_elastic_sample_update_converter import BoldSampleToElasticSampleUpdateConverter  # noqa F401
from .buffering_converter import BufferingConverter  # noqa F401
from .elastic_object_to_portaldb_object_converter import ElasticObjectToPortaldbObjectConverter  # noqa F401
from .elastic_sample_to_benchling_tissue_converter import ElasticSampleToBenchlingTissueConverter  # noqa F401
from .elastic_sample_to_benchling_tissue_update_converter import ElasticSampleToBenchlingTissueUpdateConverter  # noqa F401
from .elastic_sample_to_bold_sample_converter import ElasticSampleToBoldSampleConverter  # noqa F401
from .elastic_sample_to_elastic_sequencing_request_update_converter import ElasticSampleToElasticSequencingRequestUpdateConverter  # noqa F401
from .elastic_sample_to_sts_sample_converter import ElasticSampleToStsSampleConverter  # noqa F401
from .elastic_sequencing_request_to_sts_sequencing_request_converter import ElasticSequencingRequestToStsSequencingRequestConverter  # noqa F401
from .elastic_sequencing_request_to_elastic_run_data_update_converter import ElasticSequencingRequestToElasticRunDataUpdateConverter  # noqa F401
from .elastic_species_to_sts_species_converter import ElasticSpeciesToStsSpeciesConverter  # noqa F401
from .elastic_tolid_to_elastic_curation_update_converter import ElasticTolidToElasticCurationUpdateConverter  # noqa F401
from .elastic_tolid_to_elastic_genome_note_update_converter import ElasticTolidToElasticGenomeNoteUpdateConverter  # noqa F401
from .elastic_tolid_to_elastic_sample_update_converter import ElasticTolidToElasticSampleUpdateConverter  # noqa F401
from .gap_assembly_to_elastic_assembly_converter import GapAssemblyToElasticAssemblyConverter  # noqa F401
from .gap_assembly_to_elastic_assembly_analysis_converter import GapAssemblyToElasticAssemblyAnalysisConverter  # noqa F401
from .genome_notes_genome_note_to_elastic_genome_note_converter import GenomeNotesGenomeNoteToElasticGenomeNoteConverter  # noqa F401
from .goat_taxon_to_elastic_species_converter import GoatTaxonToElasticSpeciesConverter  # noqa F401
from .grit_issue_to_elastic_curation_converter import GritIssueToElasticCurationConverter  # noqa F401
from .grit_issue_to_elastic_issue_converter import GritIssueToElasticIssueConverter  # noqa F401
from .incoming_sample_to_ena_sample_converter import IncomingSampleToEnaSampleConverter # noqa
from .incoming_sample_to_incoming_sample_with_lists_converter import IncomingSampleToIncomingSampleWithListsConverter  # noqa F401
from .informatics_tolid_to_elastic_tolid_converter import InformaticsTolidToElasticTolidConverter  # noqa F401
from .labwhere_location_to_elastic_sample_update_converter import LabwhereLocationToElasticSampleUpdateConverter  # noqa F401
from .labwhere_location_to_sts_tray_converter import LabwhereLocationToStsTrayConverter  # noqa F401
from .mlwh_extraction_to_elastic_extraction_converter import MlwhExtractionToElasticExtractionConverter  # noqa F401
from .mlwh_run_data_to_elastic_run_data_converter import MlwhRunDataToElasticRunDataConverter  # noqa F401
from .mlwh_sequencing_request_to_elastic_sequencing_request_converter import MlwhSequencingRequestToElasticSequencingRequestConverter  # noqa F401
from .sts_banked_sample_to_elastic_sample_converter import StsBankedSampleToElasticSampleConverter  # noqa F401
from .sts_manifest_to_elastic_manifest_converter import StsManifestToElasticManifestConverter  # noqa F401
from .sts_project_to_elastic_sample_update_converter import StsProjectToElasticSampleUpdateConverter  # noqa F401
from .sts_sample_project_to_elastic_sample_converter import StsSampleProjectToElasticSampleConverter  # noqa F401
from .sts_sampleset_to_elastic_sampleset_converter import StsSamplesetToElasticSamplesetConverter  # noqa F401
from .sts_species_to_elastic_species_converter import StsSpeciesToElasticSpeciesConverter  # noqa F401
from .tolid_specimen_to_elastic_tolid_converter import TolidSpecimenToElasticTolidConverter  # noqa F401
from .tolqc_data_to_elastic_run_data_converter import TolqcDataToElasticRunDataConverter  # noqa F401
from .tolqc_sample_to_elastic_sequencing_request_converter import TolqcSampleToElasticSequencingRequestConverter  # noqa F401
from .tolqc_species_to_elastic_species_converter import TolqcSpeciesToElasticSpeciesConverter  # noqa F401
from .sts_sample_to_casm_benchling_converter import StsSampleToCasmBenchlingConverterFactory # noqa F401
from .treeofsex_species_to_treeofsexwh_species_converter import TreeofsexSpeciesToTreeofsexwhSpeciesConverter  # noqa F401
from .treeofsex_upload_to_treeofsex_attribute_converter import TreeofsexUploadToTreeofsexAttributeConverter  # noqa F401
from .treeofsex_upload_to_treeofsexwh_species_converter import TreeofsexUploadToTreeofsexwhSpeciesConverter  # noqa F401
from .skip_null_fields_converter import SkipNullFieldsConverter  # noqa F401
from .default_field_value_if_missing_converter import DefaultFieldValueIfMissingConverter  # noqa F401
from .prefix_field_converter import PrefixFieldConverter  # noqa F401
from .combine_fields_converter import CombineFieldsConverter  # noqa F401
from .auto_detect_manifest_type_converter import AutoDetectManifestTypeConverter # noqa F401
from .time_string_to_time import TimeStringToTimeConverter # noqa F401


# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .parser import Parser  # noqa F401
from .converter import S3Converter  # noqa F401
from .factory import create_s3_datasource  # noqa F401
from .s3_datasource import S3DataSource  # noqa F401



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from datetime import datetime, time

from tol.core import DataObject
from tol.core.validate import Validator


class TypesValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances,
    ensuring that they only have attributes of the given
    allowed keys.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        allowed_types: dict[str, str]
        is_error: bool = True
        detail: str = 'Value is of incorrect type'

    __slots__ = ['__config']
    __config: Config

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        super().__init__()
        self.__config = config

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:

        type_map = {
            'str': str,
            'int': int,
            'float': float,
            'bool': bool,
            'list': list,
            'dict': dict,
            'datetime': datetime,
            'time': time
        }
        for key, expected_type in self.__config.allowed_types.items():
            if key in obj.attributes:
                actual_value = obj.get_field_by_name(key)
                if actual_value is None:
                    continue
                type_class = type_map.get(expected_type)
                if type_class and not isinstance(actual_value, type_class):
                    if expected_type == 'time':
                        self.__add_result(
                            obj,
                            key,
                            detail='Only HH:MM format is accepted.',
                        )
                    else:
                        self.__add_result(
                            obj,
                            key,
                            detail=f'Field {key} value "{actual_value}" is not of type '
                                   f'"{expected_type}"',
                        )
                if type_class and isinstance(actual_value, type_class):
                    # Special case for bool since isinstance(True, int) is True
                    if expected_type == 'int' and isinstance(actual_value, bool):
                        self.__add_result(
                            obj,
                            key,
                        )

    def __add_result(
        self,
        obj: DataObject,
        key: str,
        detail: str = None,
    ) -> None:

        if self.__config.is_error:
            self.add_error(
                object_id=obj.id,
                detail=detail or self.__config.detail,
                field=key,
            )
        else:
            self.add_warning(
                object_id=obj.id,
                detail=self.__config.detail,
                field=key,
            )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .action_parameters import ActionParameters  # noqa
from .aggregation_parameters import AggregationParameters  # noqa
from .aggregation_body import AggregationBody  # noqa
from .authenticate import Authenticator, quick_and_dirty_auth  # noqa
from .auth_context import AuthContext, CtxGetter, default_ctx_getter  # noqa
from .data_body import RequestBody, JsonApiRequestBody  # noqa
from .filter_utils import FilterUtils  # noqa
from .list_get_parameters import ListGetParameters  # noqa
from .relation_url import RelataionshipHopsParser  # noqa
from .stats_parameters import GroupStatsParameters, StatsParameters  # noqa
